import argparse

import matplotlib.pyplot as plt
import pandas as pd

from go_utils import lc, mhm
from go_utils.download import convert_dates_to_datetime, get_api_data
from go_utils.geoenrich import get_country_api_data
from go_utils.photo_download import download_lc_photos, download_mhm_photos

protocol_map = {"mosquito": "mosquito_habitat_mapper", "landcover": "land_covers"}


def add_download_args(parser):
    parser.add_argument("--out", "-o", help="Output Directory of the command")
    parser.add_argument(
        "--start",
        "-s",
        help="Start date (if you want data from the api, so don't specify -i)",
    )
    parser.add_argument(
        "--end",
        "-e",
        help="End date (if you want data from the api, so don't specify -i)",
    )
    parser.add_argument(
        "--countries",
        "-co",
        help="Desired Countries separated by commas (if you want data from the api, so don't specify -i)",
    )
    parser.add_argument(
        "--regions",
        "-r",
        help="Desired Regions separated by commas (if you want data from the api, so don't specify -i)",
    )
    parser.add_argument(
        "--box",
        "-b",
        help="Bounding Box (if you want data from the api, so don't specify -i). Put coordinates in order of 'min lat, min lon, max lat, max lon'",
        type=list,
    )


def download_data(protocol, args):
    func_args = {"protocol": protocol}
    if args.start:
        func_args["start_date"] = args.start
    if args.end:
        func_args["end_date"] = args.end
    if args.box:
        coords = [float(coord.strip()) for coord in args.box.split(",")]
        box = {
            "min_lat": coords[0],
            "min_lon": coords[1],
            "max_lat": coords[1],
            "max_lon": coords[3],
        }
        func_args["latlon_box"] = box
    if args.countries:
        func_args["countries"] = [
            country.strip() for country in args.countries.split(",")
        ]
    if args.regions:
        func_args["regions"] = [region.strip() for region in args.regions.split(",")]

    if "countries" in func_args or "regions" in func_args:
        df = get_country_api_data(**func_args)
    else:
        df = get_api_data(**func_args)

    return df


def mhm_data_download():
    parser = argparse.ArgumentParser(
        description="GLOBE Observer Mosquito Habitat Mapper API Download CLI"
    )
    add_download_args(parser)
    parser.add_argument(
        "--hasgenus",
        "-hg",
        help="Filter data if it has a genus as a record",
        action="store_true",
    )
    parser.add_argument(
        "--iscontainer",
        "-ic",
        help="Filter data if the record's watersource is a container",
        action="store_true",
    )
    parser.add_argument(
        "--hasphotos",
        "-hp",
        help="Filter data if it has photo records",
        action="store_true",
    )
    parser.add_argument(
        "--minlarvae", "-ml", help="Filter data by minimum larvae count", type=int
    )
    args = parser.parse_args()
    df = download_data("mosquito_habitat_mapper", args)

    filter_args = {}
    filter_args["has_genus"] = args.hasgenus
    filter_args["is_container"] = args.iscontainer
    filter_args["has_photos"] = args.hasphotos

    if args.minlarvae:
        filter_args["min_larvae_count"] = args.minlarvae

    df = mhm.qa_filter(df, **filter_args)

    if args.out:
        df.to_csv(args.out)
    else:
        mhm.diagnostic_plots(df)
        plt.show()


def lc_data_download():
    parser = argparse.ArgumentParser(
        description="GLOBE Observer Land Cover API Download CLI"
    )
    add_download_args(parser)

    # TODO: Group args to reduce user confusion
    parser.add_argument(
        "--hasclassification",
        "-hc",
        help="Filter data if it has atleast one classification as a record",
        action="store_true",
    )
    parser.add_argument(
        "--hasphoto",
        "-hp",
        help="Filter data if it has atleast one photo as a record",
        action="store_true",
    )
    parser.add_argument(
        "--hasallclassifications",
        "-hac",
        help="Filter data if it has all classifications for each record",
        action="store_true",
    )
    parser.add_argument(
        "--hasallphotos",
        "-hap",
        help="Filter data if it has all photos for each record",
        action="store_true",
    )
    args = parser.parse_args()
    df = download_data("land_covers", args)

    filter_args = {}
    filter_args["has_classification"] = args.hasclassification
    filter_args["has_photo"] = args.hasphoto
    filter_args["has_all_classifications"] = args.hasallclassifications
    filter_args["has_all_photos"] = args.hasallphotos
    df = lc.qa_filter(df, **filter_args)

    if args.out:
        df.to_csv(args.out)
    else:
        lc.diagnostic_plots(df)
        plt.show()


def mhm_photo_download():
    parser = argparse.ArgumentParser(
        description="GLOBE Observer Mosquito Habitat Mapper Photo Downloader CLI"
    )
    parser.add_argument("input", help="Input Directory of the command")
    parser.add_argument("out", help="Output Directory of the command")

    photo_group = parser.add_argument_group(
        "Photo Types", "Includes: larvae, watersource, abdomen, all"
    )
    photo_group.add_argument(
        "--larvae", "-l", help="Include Larvae Photos", action="store_true"
    )
    photo_group.add_argument(
        "--watersource", "-w", help="Include Watersource Photos", action="store_true"
    )
    photo_group.add_argument(
        "--abdomen", "-ab", help="Include Abdomen Photos", action="store_true"
    )

    photo_group.add_argument(
        "--all",
        "-a",
        help="Include all of the flags",
        action="store_true",
    )

    name_group = parser.add_argument_group(
        "Photo Naming", "Allows naming customization"
    )
    name_group.add_argument(
        "--name_additional",
        "-add",
        help="Include additonal info in photo names.",
        type=str,
        action="store",
    )

    mhm_name_fields = [
        "url_type",
        "watersource",
        "latitude",
        "longitude",
        "date_str",
        "mhm_id",
        "classification",
    ]
    name_group.add_argument(
        "--nargs_include",
        "-in",
        nargs="+",
        help="Include these fields from photo names.",
        choices=mhm_name_fields,
        type=str,
        action="store",
    )

    args = parser.parse_args()

    download_args = {}
    if not args.larvae:
        download_args["larvae_photo"] = ""
    if not args.watersource:
        download_args["watersource_photo"] = ""
    if not args.abdomen:
        download_args["abdomen_photo"] = ""

    df = pd.read_csv(args.input)
    convert_dates_to_datetime(df)

    if args.all:
        download_mhm_photos(
            df,
            args.out,
            include_in_name=args.nargs_include,
            additional_name_stem=args.name_additional,
        )
    else:
        download_mhm_photos(
            df,
            args.out,
            **download_args,
            include_in_name=args.nargs_include,
            additional_name_stem=args.name_additional
        )


def lc_photo_download():
    parser = argparse.ArgumentParser(
        description="GLOBE Observer Land Cover Photo Downloader CLI"
    )
    parser.add_argument("input", help="Input Directory of the command")
    parser.add_argument("out", help="Output Directory of the command")

    photo_group = parser.add_argument_group(
        "Photo Types", "Includes: up, down, north, south, east, west"
    )
    photo_group.add_argument(
        "--up", "-u", help="Include Upward Photos", action="store_true"
    )
    photo_group.add_argument(
        "--down", "-d", help="Include Downward Photos", action="store_true"
    )
    photo_group.add_argument(
        "--north", "-n", help="Include Northern Photos", action="store_true"
    )
    photo_group.add_argument(
        "--south", "-s", help="Include Southern Photos", action="store_true"
    )
    photo_group.add_argument(
        "--east", "-e", help="Include Eastern Photos", action="store_true"
    )
    photo_group.add_argument(
        "--west", "-w", help="Include Western Photos", action="store_true"
    )
    photo_group.add_argument(
        "--all", "-a", help="Include All Photos", action="store_true"
    )

    name_group = parser.add_argument_group(
        "Photo Naming", "Allows naming customization"
    )
    name_group.add_argument(
        "--name_additional",
        "-add",
        help="Include additonal info in photo names.",
        type=str,
        action="store",
    )

    lc_name_fields = ["direction", "latitude", "longitude", "date_str", "lc_id"]
    name_group.add_argument(
        "--nargs_include",
        "-in",
        nargs="+",
        help="Exclude these fields from photo names.",
        choices=lc_name_fields,
        type=str,
        action="store",
    )

    args = parser.parse_args()

    download_args = {}
    if not args.up:
        download_args["up_photo"] = ""
    if not args.down:
        download_args["down_photo"] = ""
    if not args.north:
        download_args["north_photo"] = ""
    if not args.south:
        download_args["south_photo"] = ""
    if not args.east:
        download_args["east_photo"] = ""
    if not args.west:
        download_args["west_photo"] = ""

    df = pd.read_csv(args.input)
    convert_dates_to_datetime(df)

    if args.all:
        download_lc_photos(
            df,
            args.out,
            include_in_name=args.nargs_include,
            additional_name_stem=args.name_additional,
        )
    else:
        download_lc_photos(
            df,
            args.out,
            **download_args,
            include_in_name=args.nargs_include,
            additional_name_stem=args.name_additional
        )



import numpy as np
from pandas.api.types import is_hashable

__doc__ = """
# Overview
This submodule contains code to facilitate the general filtering of data. 
The following sections discuss some of the logic and context behind these methods.

# Methods

## Filter Invalid Coords
Certain entries in the GLOBE Database have latitudes and longitudes that don't exist.

## Filter Duplicates
Due to reasons like GLOBE Observer trainings among other things, there are oftentimes multiple observations of the same exact entry. This can lead to a decrease in data quality and so this utility can be used to reduce this. Groups of entries that share the same MGRS Latitude, NGRS Longitude, measured date, and other dataset specific attributes (e.g. water source) could likely be duplicate entries. In (Low, et. al)[https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2021GH000436], Mosquito Habitat Mapper duplicates are removed by groups of size greater than 10 sharing MGRS Latitude, MGRS Longitude, Water source, and Sitename values.

## Filter Poor Geolocational Data
Geolocational data may not be the most accurate. As a result, this runs a relatively naive check to remove poor geolocational data. More specifically, if the MGRS coordinates match up with the GPS coordinates or the GPS coordinates are whole numbers, then the entry is considered poor quality.
"""


def filter_out_entries(df, mask, include, inplace):
    """
    Filters out or selects target entries of a DataFrame using a mask. Mainly serves as a utility function for the other filters.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to filter
    mask : 1D np.array of bools
        The mask to apply to the DataFrame
    include : bool
        True to only select the masked values False to exclude the masked values
    inplace : bool
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the mask filter applied. If `inplace=True` it returns None.
    """
    if include:
        final_mask = mask
    else:
        final_mask = ~mask
    filtered_df = df[final_mask]
    if not inplace:
        return filtered_df
    else:
        df.mask(~df.isin(filtered_df), inplace=True)
        df.dropna(how="all", inplace=True)
        for col in df.columns:
            if df[col].dtype != filtered_df[col].dtype:
                df[col] = df[col].astype(filtered_df[col].dtype)


def filter_invalid_coords(
    df, latitude_col, longitude_col, inclusive=False, inplace=False
):
    """
    Filters latitude and longitude of a DataFrame to lie within the latitude range of [-90, 90] or (-90, 90) and longitude range of [-180, 180] or (-180, 180)

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to filter
    latitude_col : str
        The name of the column that contains latitude values
    longitude_col : str
        The name of the column that contains longitude values
    inclusive : bool, default=False
        True if you would like the bounds of the latitude and longitude to be inclusive e.g. [-90, 90]. Do note that these bounds may not work with certain GIS software and projections.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with invalid latitude and longitude entries removed. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()

    if inclusive:
        mask = (
            (df[latitude_col] >= -90)
            & (df[latitude_col] <= 90)
            & (df[longitude_col] <= 180)
            & (df[longitude_col] >= -180)
        )
    else:
        mask = (
            (df[latitude_col] > -90)
            & (df[latitude_col] < 90)
            & (df[longitude_col] < 180)
            & (df[longitude_col] > -180)
        )

    return filter_out_entries(df, mask, True, inplace)


def filter_duplicates(df, columns, group_size, inplace=False):
    """
    Filters possible duplicate data by grouping together suspiciously similar entries.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to filter
    columns : list of str
        The name of the columns that duplicate data would share. This can include things such as MGRS Latitude, MGRS Longitude, measure date, and other fields (e.g. mosquito water source for mosquito habitat mapper).
    group_size : int
        The number of duplicate entries in a group needed to classify the group as duplicate data.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with duplicate data removed. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()

    # groups / filters suspected events
    suspect_df = df.groupby(by=columns).filter(lambda x: len(x) >= group_size)
    suspect_mask = df.isin(suspect_df)
    suspect_mask = np.any(suspect_mask, axis=1)

    return filter_out_entries(df, suspect_mask, False, inplace)


def filter_poor_geolocational_data(
    df,
    latitude_col,
    longitude_col,
    mgrs_latitude_col,
    mgrs_longitude_col,
    inplace=False,
):
    """
    Filters latitude and longitude of a DataFrame that contain poor geolocational quality.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to filter
    latitude_col : str
        The name of the column that contains latitude values
    longitude_col : str
        The name of the column that contains longitude values
    mgrs_latitude_col : str
        The name of the column that contains MGRS latitude values
    mgrs_longitude_col : str
        The name of the column that contains MGRS longitude values
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with bad latitude and longitude entries removed. If `inplace=True` it returns None.
    """

    def geolocational_filter(gps_lat, gps_lon, recorded_lat, recorded_lon):
        return (
            (recorded_lat == gps_lat and recorded_lon == gps_lon)
            or gps_lat == int(gps_lat)
            or gps_lon == int(gps_lon)
        )

    if not inplace:
        df = df.copy()

    vectorized_filter = np.vectorize(geolocational_filter)
    bad_data = vectorized_filter(
        df[latitude_col].to_numpy(),
        df[longitude_col].to_numpy(),
        df[mgrs_latitude_col].to_numpy(),
        df[mgrs_longitude_col].to_numpy(),
    )

    return filter_out_entries(df, bad_data, False, inplace)


def filter_by_globe_team(
    df, globe_teams_column, target_teams, exclude=False, inplace=False
):
    """
    Finds or filters out specific globe teams.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to filter
    globe_teams_column : str
        The column containing the GLOBE teams.
    target_teams : list of str
        The names of the GLOBE teams to be used.
    exclude : bool, default=False
        Whether to exclude the specified teams from the dataset.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.
    Returns
    -------
    pd.DataFrame or None
        A DataFrame with only the specified GLOBE teams (if exclude is False) or without the specified GLOBE teams (if exclude is True). If `inplace=True` it returns None.
    """

    def is_desired_team(team_list):
        if not exclude:
            return any(
                [
                    team in team_list if not is_hashable(team_list) else False
                    for team in target_teams
                ]
            )
        else:
            return all(
                [
                    team not in team_list if not is_hashable(team_list) else False
                    for team in target_teams
                ]
            )

    desired_team_filter = np.vectorize(is_desired_team)
    desired_data_mask = desired_team_filter(df[globe_teams_column].to_numpy())

    return filter_out_entries(df, desired_data_mask, True, inplace)



import logging

import numpy as np
import pandas as pd
from pytz import timezone
from timezonefinder import TimezoneFinder

__doc__ = """

# Overview
This submodule contains several methods to assist with data cleanup.
The following sections discuss some of the decisions behind these methods and their part of a larger data cleanup pipeline.

# Methods

## Remove Redundant/Homogenous Column:
[This method](#remove_homogenous_cols) indentifies and removes columns where all values were the same. If the logging level is set to INFO, the method will also print the names of the dropped columns and their respective singular value as a Python dictionary. 
For the raw mosquito habitat mapper data, the following were dropped: 
- {'protocol': 'mosquito_habitat_mapper'} 
- {'ExtraData': None} 
- {'MosquitoEggCount': None} 
- {'DataSource': 'GLOBE Observer App'}

For raw landcover data, the following were dropped:
- {'protocol': 'land_covers'}

## Rename columns:

### Differentiating between MGRS and GPS Columns: 
The GLOBE API data for `MosquitoHabitatMapper` and `LandCovers` report each observation’s Military Grid Reference System (MGRS) Coordinates in the `latitude` and `longitude` fields. The GPS Coordinates are stored in the `MeasurementLatitude` and `MeasurementLongitude` fields. 
To avoid confusion between these measuring systems, [this method](#rename_latlon_cols) renames `latitude` and `longitude` to `MGRSLatitude` and `MGRSLongitude`, respectively, and `MeasurementLatitude` and `MeasurementLongitude` to `Latitude` and `Longitude`, respectively. Now, the official `Latitude` and `Longitude` columns are more intuitively named.

### Protocol Abbreviation:
To better support future cross-protocol analysis and data enrichment, [this method](#replace_column_prefix) following naming scheme for all column names: `protocolAbbreviation_columnName`, where `protocolAbbreviation` was the abbreviation for the protocol (`mhm` for mosquito habitat mapper and `lc` for land cover) and `columnName` was the original name of the column. For example, the mosquito habitat mapper “MGRSLongitude” column was renamed “mhm_MGRSLongitude” and the corresponding land cover column was renamed “lc_MGRSLongitude”.

Do note that if you would like to use the previously mentioned `mhm` and `lc` naming scheme for you data, the `go_utils.mhm` and `go_utils.lc` submodules each have a method called `cleanup_column_prefix` which uses the previously mentioned naming scheme as opposed to the `replace_column_prefix` method which requires that you specify the current prefix and desired prefix.

## Standardize no-data values:
The GLOBE API CSV’s lacked standardization in indicating No Data. Indicators ranged from Python's `None`, to `“null”`, to an empty cell, to `NaN` (`np.nan`). To improve the computational efficiency in future mathematical algorithms on the GLOBE datasets, [this method](#standardize_null_values) converts all No Data Indicators to np.nan (Python NumPy’s version of No-Data as a float). Do note that later in Round Appropriate Columns, all numerical extraneous values are converted from np.nan to -9999.  Thus, Users will receive the pre-processed GLOBE API Mosquito Habitat Mapper and Land Cover Data in accordance with the standards described by Cook et al (2018). 

## Round Appropriate Columns
[This method](#round_cols) does the following:
1. Identifies all numerical columns (e.g. `float64`, `float`, `int`, `int64`).
2. Rounds Latitudes and Longitude Columns to 5 places. To reduce data density, all latitude and longitude values were rounded to 5 decimal places. This corresponds to about a meter of accuracy. Furthermore, any larger number of decimal places consume unnecessary amounts of storage as the GLOBE Observer app cannot attain such precision.
3. Converts other Numerical Data to Integers. To improve the datasets’ memory and performance, non latitude and longitude numerical values were converted to integers for the remaining columns, including `Id`, `MeasurementElevation`, and `elevation` columns.  This is appropriate since ids are always discrete values. `MeasurementElevation` and `elevation` are imprecise estimates from 3rd party sources, rendering additional precision an unnecessary waste of memory. However, by converting these values to integers, we could no longer use np.nan, a float, to denote extraneous/empty values. Thus, for integer columns, we used -9999 to denote extraneous/empty values.

**Note**: Larvae Counts were also converted to integers and Land Classification Column percentages were also converted to integers, reducing our data density. This logic is further discussed in go_utils.mhm.larvae_to_num for mosquito habitat mapper and go_utils.lc.unpack_classifications

"""


def adjust_timezones(df, time_col, latitude_col, longitude_col, inplace=False):
    """
    Calculates timezone offset and adjusts date columns accordingly. This is done because GLOBE data uses UTC timezones and it can be useful to have the time adjusted to the local observation time.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to adjust time zones for
    time_col : str
        The column that contains the time zone data
    latitude_col : str
        The column that contains latitude data
    longitude_col : str
        The column that contains longitude data
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with its time entry adjusted to its local timezone. If `inplace=True` it returns None.
    """
    tf = TimezoneFinder()

    def convert_timezone(time, latitude, longitude):
        utc_tz = pd.to_datetime(time, utc=True)
        local_time_zone = timezone(tf.timezone_at(lng=longitude, lat=latitude))
        return utc_tz.astimezone(local_time_zone)

    time_zone_converter = np.vectorize(convert_timezone)

    if not inplace:
        df = df.copy()

    df[time_col] = time_zone_converter(
        df[time_col].to_numpy(),
        df[latitude_col].to_numpy(),
        df[longitude_col].to_numpy(),
    )

    if not inplace:
        return df


def remove_homogenous_cols(df, exclude=[], inplace=False):
    """
    Removes columns froma DataFrame if they contain only 1 unique value.
    ```

    Then the original `df` variable that was passed is now updated with these dropped columns.

    If you would like to see the columns that are dropped, setting the logging level to info will allow for that to happen.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame that will be modified
    exclude : list of str, default=[]
        A list of any columns that should be excluded from this removal.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with homogenous columns removed. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()

    for column in df.columns:
        try:
            if column not in exclude and len(pd.unique(df[column])) == 1:
                logging.info(f"Dropped: {df[column].iloc[0]}")
                df.drop(column, axis=1, inplace=True)
        except TypeError:
            continue

    if not inplace:
        return df


def replace_column_prefix(df, current_prefix, replacement_text, inplace=False):
    """
    Replaces the protocol prefix (e.g. mosquito_habitat_mapper/mosquitohabitatmapper) for the column names with another prefix in the format of `newPrefix_columnName`.

    If you are interested in replacing the prefixes for the raw mosquito habitat mapper and landcover datasets, use the go_utils.lc.cleanup_column_prefix and go_utils.mhm.cleanup_column_prefix methods.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame you would like updated
    protocol : str
        A string representing the protocol prefix.
    replacement_text : str
        A string representing the desired prefix for the column name.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the column prefixes replaced. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()
    df.columns = [
        f"{replacement_text}_{column.replace(current_prefix,'')}"
        for column in df.columns
    ]
    if not inplace:
        return df


def find_column(df, keyword):
    """Finds the first column that contains a certain keyword. Mainly intended to be a utility function for some of the other methods.


    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing the columns that need to be searched.
    keyword: str
        The keyword that needs to be present in the desired column.
    """

    return [column for column in df.columns if keyword in column][0]


def camel_case(string, delimiters=[" "]):
    """Converts a string into camel case

    Parameters
    ----------
    string: str, the string to convert
    delimiter: str, the character that denotes separate words
    """
    for delimiter in delimiters:
        str_list = [s[0].upper() + s[1:] for s in string.split(delimiter)]
        string = "".join([s for s in str_list])
    return string


def rename_latlon_cols(
    df,
    gps_latitude="",
    gps_longitude="",
    mgrs_latitude="latitude",
    mgrs_longitude="longitude",
    inplace=False,
):
    """Renames the latitude and longitude columns of **raw** GLOBE Observer Data to make the naming intuitive.

    [This](#differentiating-between-mgrs-and-gps-columns) explains the motivation behind the method.

    Example usage:
    ```python
    from go_utils.cleanup import rename_latlon_cols
    rename_latlon_cols(df)
    ```

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame whose columns require renaming.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the updated Latitude and Longitude column names. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()

    if not gps_latitude:
        gps_latitude = find_column(df, "MeasurementLatitude")
    if not gps_longitude:
        gps_longitude = find_column(df, "MeasurementLongitude")
    df.rename(
        {
            gps_latitude: "Latitude",
            gps_longitude: "Longitude",
            mgrs_latitude: "MGRSLatitude",
            mgrs_longitude: "MGRSLongitude",
        },
        axis=1,
        inplace=True,
    )

    if not inplace:
        return df


def round_cols(df, inplace=False):
    """This rounds columns in the DataFrame. More specifically, latitude and longitude data is rounded to 5 decimal places, other fields are rounded to integers, and null values (for the integer columns) are set to -9999.

    See [here](#round-appropriate-columns) for more information.

    Example usage:
    ```python
    from go_utils.cleanup import round_cols
    round_cols(df)
    ```

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame that requires rounding.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the rounded values. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()
    # Identifies all the numerical cols
    number_cols = [
        df.columns[i]
        for i in range(len(df.dtypes))
        if (df.dtypes[i] == "float64")
        or (df.dtypes[i] == "float")
        or (df.dtypes[i] == "int")
        or (df.dtypes[i] == "int64")
    ]

    # Rounds cols appropriately
    column_round = np.vectorize(lambda x, digits: round(x, digits))
    for name in number_cols:
        df[name] = df[name].fillna(-9999)
        if ("latitude" in name.lower()) or ("longitude" in name.lower()):
            logging.info(f"Rounded to 5 decimals: {name}")
            df[name] = column_round(df[name].to_numpy(), 5)
        else:
            logging.info(f"Converted to integer: {name}")
            df[name] = df[name].to_numpy().astype(int)

    if not inplace:
        return df


def standardize_null_vals(df, null_val=np.nan, inplace=False):
    """
    This method standardizes the null values of **raw** GLOBE Observer Data.

    ```python
    from go_utils.cleanup import standardize_null_vals
    standardize_null_vals(df)
    ```

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame that needs null value standardization
    null_val : obj, default=np.nan
        The value that all null values should be set to
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the standardized null values. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()

    # Replace Null Values with null_val
    df.fillna(null_val, inplace=True)

    # Replace any text null values
    df.replace(
        {
            "null": null_val,
            "": null_val,
            "NaN": null_val,
            "nan": null_val,
            None: null_val,
        },
        inplace=True,
    )

    if not inplace:
        return df



from datetime import datetime

import numpy as np
from arcgis.features import GeoAccessor
from arcgis.gis import GIS

from go_utils.constants import (
    abbreviation_dict,
    end_date,
    landcover_protocol,
    mosquito_protocol,
    region_dict,
    start_date,
)
from go_utils.download import convert_dates_to_datetime, default_data_clean


def get_country_api_data(
    protocol,
    start_date=start_date,
    end_date=end_date,
    is_clean=True,
    countries=[],
    regions=[],
):
    """
    Gets country enriched API Data. Due note that this data comes from layers in ArcGIS that are updated daily. Therefore, there will be some delay between when an entry is uploaded onto the GLOBE data base and being on the ArcGIS dataset.

    Parameters
    ----------
    protocol : str, {"mosquito_habitat_mapper", "land_covers"}
        The desired GLOBE Observer Protocol. Currently only mosquito habitat mapper and land cover is supported.
    start_date : str, default= 2017-05-31
        The desired start date of the dataset in the format of (YYYY-MM-DD).
    end_date : str, default= today's date in YYYY-MM-DD form.
        The desired end date of the dataset in the format of (YYYY-MM-DD).
    countries : list of str, default=[]
        The list of desired countries. Look at go_utils.info.region_dict to see supported country names. If the list is empty, all data will be included.
    regions : list of str, default=[]
        The list of desired regions. Look at go_utils.info.region_dict to see supported region names and the countries they enclose. If the list is empty, all data will be included.
    latlon_box : dict of {str, double}, optional
        The longitudes and latitudes of a bounding box for the dataset. The minimum/maximum latitudes and longitudes must be specified with the following keys: "min_lat", "min_lon", "max_lat", "max_lon". The default value specifies all latitude and longitude coordinates.
    """

    item_id_dict = {
        mosquito_protocol: "a018521fbc3f42bc848d3fa4c52e02ce",
        landcover_protocol: "fe54b831415f44d2b1640327ae276fb8",
    }

    if protocol not in item_id_dict:
        raise ValueError(
            "Invalid protocol, currently only 'mosquito_habitat_mapper' and 'land_covers' are supported."
        )

    gis = GIS()
    item = gis.content.get(itemid=item_id_dict[protocol])
    df = GeoAccessor.from_layer(item.layers[0])

    if "SHAPE" in df:
        df.drop(["SHAPE"], axis=1, inplace=True)

    # Due to the size of the mhm column names, ArcGIS truncates the names so it must be renamed in this step.
    if protocol == "mosquito_habitat_mapper":

        mhm_rename_dict = {
            col: f"mosquitohabitatmapper{col}"
            for col in df.columns
            if col[0].isupper() and col != "COUNTRY"
        }
        df.rename(
            mhm_rename_dict,
            axis=1,
            inplace=True,
        )

    # Filter the dates
    start = datetime.strptime(start_date, "%Y-%m-%d")
    end = datetime.strptime(end_date, "%Y-%m-%d")
    measured_at = protocol.replace("_", "") + "MeasuredAt"

    convert_dates_to_datetime(df)

    df = df[(df[measured_at] >= start) & (df[measured_at] <= end)]

    if is_clean:
        df = default_data_clean(df, protocol)

    for region in regions:
        countries.extend(region_dict[region])
    countries_set = set(countries)
    # Return the regular data if nothing is specified
    if not countries_set:
        return df
    else:
        mask = _get_valid_countries_mask(df, protocol, countries_set)
        return df[mask]


def _get_valid_countries_mask(df, protocol, country_list):
    country_filter = np.vectorize(lambda country_col: country_col in country_list)
    mask = country_filter(df[f"{abbreviation_dict[protocol]}_COUNTRY"].to_numpy())
    return mask



import math
import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from go_utils.cleanup import (
    camel_case,
    remove_homogenous_cols,
    rename_latlon_cols,
    replace_column_prefix,
    round_cols,
    standardize_null_vals,
)
from go_utils.plot import completeness_histogram, multiple_bar_graph, plot_freq_bar

__doc__ = """

## Unpacking the Landcover Classification Data
The classification data for each entry is condensed into several entries separated by a semicolon. [This method](#unpack_classifications) identifies and parses Land Cover Classifications and percentages to create new columns. The columns are also reordered to better group directional information together.

The end result is a DataFrame that contains columns for every Unique Landcover Classification (per direction) and its respective percentages for each entry.

There are four main steps to this procedure:
1.Identifying Land Cover Classifications for each Cardinal Direction: An internal method returns the unique description (e.g. HerbaceousGrasslandTallGrass) listed in a column. This method is run for all 4 cardinal directions to obtain the all unique classifications per direction.
2. Creating empty columns for each Classification from each Cardinal Direction: Using the newly identified classifications new columns are made for each unique classification. These columns initially contained the default float64 value of 0.0. By initializing all the classification column values to 0.0, we ensure no empty values are set to -9999 in the round_cols(df) method (discussed in General Cleanup Procedures - Round Appropriate Columns). This step eases future numerical analysis.
3. Grouping and Alphabetically Sorting Directional Column Information: To better organize the DataFrame, columns containing any of the following directional substrings: "downward", "upward", "west", "east", "north", "south" (case insensitive) are identified and alphabetically sorted. Then an internal method called move_cols, specified column headers to move (direction_data_cols), and the location before the desired point of insertion, the program returns a reordered DataFrame, where all directional columns are grouped together. This greatly improves the Land Covers dataset’s organization and accessibility.
4. Adding Classification Percentages to their respective Land Cover Classification Columns - To fill in each classification column with their respective percentages, an internal method is applied to each row in the dataframe. This method iterates through each classification direction (ie “lc_EastClassifications”) and sets each identified Classification column with its respective percentage.

NOTE: After these procedures, the original directional classification columns (e.g. “lc_EastClassifications”) are not dropped.
"""

classifications = []


def cleanup_column_prefix(df, inplace=False):
    """Method for shortening raw landcover column names.

    The df object will now replace the verbose `landcovers` prefix in some of the columns with `lc_`

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing raw landcover data. The DataFrame object itself will be modified.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the cleaned up column prefixes. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()

    replace_column_prefix(df, "landcovers", "lc", inplace=True)

    if not inplace:
        return df


def extract_classification_name(entry):
    """
    Extracts the name (landcover description) of a singular landcover classification. For example in the classification of `"60% MUC 02 (b) [Trees, Closely Spaced, Deciduous - Broad Leaved]"`, the `"Trees, Closely Spaced, Deciduous - Broad Leaved"` is extracted.

    Parameters
    ----------
    entry : str
        A single landcover classification.

    Returns
    -------
    str
        The Landcover description of a classification
    """

    return re.search(r"(?<=\[).*(?=\])", entry).group()


def extract_classification_percentage(entry):
    """
    Extracts the percentage of a singular landcover classification. For example in the classification of `"60% MUC 02 (b) [Trees, Closely Spaced, Deciduous - Broad Leaved]"`, the `60` is extracted.

    Parameters
    ----------
    entry : str
        A single landcover classification.

    Returns
    -------
    float
        The percentage of a landcover classification
    """

    return float(re.search(".*(?=%)", entry).group())


def _extract_landcover_items(func, info):
    entries = info.split(";")
    return [func(entry) for entry in entries]


def extract_classifications(info):
    """Extracts the name/landcover description (see [here](#extract_classification_name) for a clearer definition) of a landcover classification entry in the GLOBE Observer Data.

    Parameters
    ----------
    info : str
        A string representing a landcover classification entry in the GLOBE Observer Datset.

    Returns
    -------
    list of str
        The different landcover classifications stored within the landcover entry.
    """
    return _extract_landcover_items(extract_classification_name, info)


def extract_percentages(info):
    """Extracts the percentages (see [here](#extract_classification_percentage) for a clearer definition) of a landcover classification in the GLOBE Observer Datset.

    Parameters
    ----------
    info : str
        A string representing a landcover classification entry in the GLOBE Observer Datset.

    Returns
    -------
    list of float
        The different landcover percentages stored within the landcover entry.
    """

    return _extract_landcover_items(extract_classification_percentage, info)


def extract_classification_dict(info):
    """Extracts the landcover descriptions and percentages of a landcover classification entry as a dictionary.

    Parameters
    ----------
    info : str
        A string representing a landcover classification entry in the GLOBE Observer Datset.

    Returns
    -------
    dict of str, float
        The landcover descriptions and percentages stored as a dict in the form: `{"description" : percentage}`.
    """

    entries = info.split(";")
    return {
        extract_classification_name(entry): extract_classification_percentage(entry)
        for entry in entries
    }


def _get_classifications_for_direction(df, direction_col_name):
    list_of_land_types = []
    for info in df[direction_col_name]:
        # Note: Sometimes info = np.nan, a float -- In that case we do NOT parse/split
        if type(info) == str:
            [
                list_of_land_types.append(camel_case(entry, [" ", ",", "-", "/"]))
                for entry in extract_classifications(info)
            ]
    return np.unique(list_of_land_types).tolist()


def _move_cols(df, cols_to_move=[], ref_col=""):
    col_names = df.columns.tolist()
    index_before_desired_loc = col_names.index(ref_col)

    cols_before_index = col_names[: index_before_desired_loc + 1]
    cols_at_index = cols_to_move

    cols_before_index = [i for i in cols_before_index if i not in cols_at_index]
    cols_after_index = [
        i for i in col_names if i not in cols_before_index + cols_at_index
    ]

    return df[cols_before_index + cols_at_index + cols_after_index]


def unpack_classifications(
    lc_df,
    north="lc_NorthClassifications",
    east="lc_EastClassifications",
    south="lc_SouthClassifications",
    west="lc_WestClassifications",
    ref_col="lc_pid",
    unpack=True,
):
    """
    Unpacks the classification data in the *raw* GLOBE Observer Landcover data. This method assumes that the columns have been renamed with accordance to the [column cleanup](#cleanup_column_prefix) method.

    This returns a copy of the dataframe.

    See [here](#unpacking-the-landcover-classification-data) for more information.

    *Note:* The returned DataFrame will have around 250 columns.

    Parameters
    ----------
    lc_df : pd.DataFrame
        A DataFrame containing Raw GLOBE Observer Landcover data that has had the column names simplified.
    north: str, default="lc_NorthClassifications"
        The name of the column which contains the North Classifications
    east: str, default="lc_EastClassifications"
        The name of the column which contains the East Classifications
    south: str, default="lc_SouthClassifications"
        The name of the column which contains the South Classifications
    west: str, default="lc_WestClassifications"
        The name of the column which contains the West Classifications
    ref_col: str, default="lc_pid"
        The name of the column which all of the expanded values will be placed after. For example, if the columns were `[1, 2, 3, 4]` and you chose 3, the new columns will now be `[1, 2, 3, (all classification columns), 4]`.
    unpack: bool, default=True
        True if you want to unpack the directional classifications, False if you only want overall classifications

    Returns
    -------
    pd.DataFrame
        A DataFrame with the unpacked classification columns.
    list
        A list containing all the generated overall Land Cover column names (mainly for testing purposes).
    list
        A list containing all the generated directional Land Cover column names (mainly for testing purposes).
    """

    classifications = [north, east, south, west]

    def set_directions(row):
        for classification in classifications:
            if not pd.isnull(row[classification]):
                entries = row[classification].split(";")
                for entry in entries:
                    percent, name = (
                        extract_classification_percentage(entry),
                        extract_classification_name(entry),
                    )
                    name = camel_case(name, [" ", ",", "-", "/"])
                    classification = classification.replace("Classifications", "_")
                    overall = re.sub(
                        r"(north|south|east|west).*",
                        "Overall_",
                        key,
                        flags=re.IGNORECASE,
                    )
                    row[f"{classification}{name.strip()}"] = percent
                    row[f"{overall}{name.strip()}"] += percent
        return row

    land_type_columns_to_add = {
        classification: _get_classifications_for_direction(lc_df, classification)
        for classification in classifications
    }
    overall_columns = set()
    direction_cols = set()
    for key, values in land_type_columns_to_add.items():
        direction_name = key.replace("Classifications", "_")
        overall = re.sub(
            r"(north|south|east|west).*", "Overall_", key, flags=re.IGNORECASE
        )
        for value in values:
            direction_cols.add(direction_name + value)
            overall_columns.add(overall + value)
    overall_columns = list(overall_columns)
    direction_cols = list(direction_cols)
    direction_data_cols = sorted(overall_columns + direction_cols)

    # Creates a blank DataFrame and concats it to the original to avoid iteratively growing the LC DataFrame
    blank_df = pd.DataFrame(
        np.zeros((len(lc_df), len(direction_data_cols))), columns=direction_data_cols
    )

    lc_df = pd.concat([lc_df, blank_df], axis=1)

    lc_df = _move_cols(lc_df, cols_to_move=direction_data_cols, ref_col=ref_col)
    lc_df = lc_df.apply(set_directions, axis=1)
    for column in overall_columns:
        lc_df[column] /= 4

    if not unpack:
        lc_df = lc_df.drop(columns=direction_cols)
    return lc_df, overall_columns, direction_cols


def photo_bit_flags(
    df,
    up="lc_UpwardPhotoUrl",
    down="lc_DownwardPhotoUrl",
    north="lc_NorthPhotoUrl",
    south="lc_SouthPhotoUrl",
    east="lc_EastPhotoUrl",
    west="lc_WestPhotoUrl",
    photo_count="lc_PhotoCount",
    rejected_count="lc_RejectedCount",
    pending_count="lc_PendingCount",
    empty_count="lc_EmptyCount",
    bit_binary="lc_PhotoBitBinary",
    bit_decimal="lc_PhotoBitDecimal",
    inplace=False,
):
    """
    Creates the following flags:
    - `PhotoCount`: The number of valid photos per record.
    - `RejectedCount`: The number of photos that were rejected per record.
    - `PendingCount`: The number of photos that are pending approval per record.
    - `PhotoBitBinary`: A string that represents the presence of a photo in the Up, Down, North, South, East, and West directions. For example, if the entry is `110100`, that indicates that there is a valid photo for the Up, Down, and South Directions but no valid photos for the North, East, and West Directions.
    - `PhotoBitDecimal`: The numerical representation of the lc_PhotoBitBinary string.

    Parameters
    ----------
    df : pd.DataFrame
        A land cover DataFrame
    up : str, default="lc_UpwardPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the upwards photo.
    down : str, default="lc_DownwardPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the downwards photo.
    north : str, default="lc_NorthPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the north photo.
    south : str, default="lc_SouthPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the south photo.
    east : str, default="lc_EastPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the east photo.
    west : str, default="lc_WestPhotoUrl"
        The name of the column in the land cover DataFrame that contains the url for the west photo.
    photo_count : str, default="lc_PhotoCount"
        The name of the column that will be storing the PhotoCount flag.
    rejected_count : str, default="lc_RejectedCount"
        The name of the column that will be storing the RejectedCount flag.
    pending_count : str, default="lc_PendingCount"
        The name of the column that will be storing the PendingCount flag.
    empty_count : str, default="lc_EmptyCount"
        The name of the column that will be storing the EmptyCount flag.
    bit_binary : str, default="lc_PhotoBitBinary"
        The name of the column that will be storing the PhotoBitBinary flag.
    bit_decimal : str, default="lc_PhotoBitDecimal"
        The name of the column that will be storing the PhotoBitDecimal flag.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the photo bit flags. If `inplace=True` it returns None.
    """

    def pic_data(*args):
        pic_count = 0
        rejected_count = 0
        pending_count = 0
        empty_count = 0
        valid_photo_bit_mask = ""

        for entry in args:
            if not pd.isna(entry) and "http" in entry:
                valid_photo_bit_mask += "1"
                pic_count += entry.count("http")
            else:
                valid_photo_bit_mask += "0"
            if pd.isna(entry):
                empty_count += 1
            else:
                pending_count += entry.count("pending")
                rejected_count += entry.count("rejected")
        return (
            pic_count,
            rejected_count,
            pending_count,
            empty_count,
            valid_photo_bit_mask,
            int(valid_photo_bit_mask, 2),
        )

    if not inplace:
        df = df.copy()

    get_photo_data = np.vectorize(pic_data)
    (
        df[photo_count],
        df[rejected_count],
        df[pending_count],
        df[empty_count],
        df[bit_binary],
        df[bit_decimal],
    ) = get_photo_data(
        df[up].to_numpy(),
        df[down].to_numpy(),
        df[north].to_numpy(),
        df[south].to_numpy(),
        df[east].to_numpy(),
        df[west].to_numpy(),
    )

    if not inplace:
        return df


def classification_bit_flags(
    df,
    north="lc_NorthClassifications",
    south="lc_SouthClassifications",
    east="lc_EastClassifications",
    west="lc_WestClassifications",
    classification_count="lc_ClassificationCount",
    bit_binary="lc_ClassificationBitBinary",
    bit_decimal="lc_ClassificationBitDecimal",
    inplace=False,
):
    """
    Creates the following flags:
    - `ClassificationCount`: The number of classifications per record.
    - `BitBinary`: A string that represents the presence of a classification in the North, South, East, and West directions. For example, if the entry is `1101`, that indicates that there is a valid classification for the North, South, and West Directions but no valid classifications for the East Direction.
    - `BitDecimal`: The number of photos that are pending approval per record.

    Parameters
    ----------
    df : pd.DataFrame
        A land cover DataFrame
    north : str, default="lc_NorthClassifications"
        The name of the column in the land cover DataFrame that contains the north classification.
    south : str, default="lc_SouthClassifications"
        The name of the column in the land cover DataFrame that contains the south classification.
    east : str, default="lc_EastClassifications"
        The name of the column in the land cover DataFrame that contains the east classification.
    west : str, default="lc_WestClassifications"
        The name of the column in the land cover DataFrame that contains the west classification.
    classification_count : str, default="lc_ClassificationCount"
        The name of the column that will store the ClassificationCount flag.
    bit_binary : str, default="lc_ClassificationBitBinary"
        The name of the column that will store the BitBinary flag.
    bit_decimal : str, default="lc_ClassificationBitDecimal"
        The name of the column that will store the BitDecimal flag.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the classification bit flags. If `inplace=True` it returns None.
    """

    def classification_data(*args):
        classification_count = 0
        classification_bit_mask = ""
        for entry in args:
            if pd.isna(entry) or entry is np.nan:
                classification_bit_mask += "0"
            else:
                classification_count += 1
                classification_bit_mask += "1"
        return (
            classification_count,
            classification_bit_mask,
            int(classification_bit_mask, 2),
        )

    if not inplace:
        df = df.copy()
    get_classification_data = np.vectorize(classification_data)

    (
        df[classification_count],
        df[bit_binary],
        df[bit_decimal],
    ) = get_classification_data(
        df[north],
        df[south],
        df[east],
        df[west],
    )
    if not inplace:
        return df


def completion_scores(
    df,
    photo_bit_binary="lc_PhotoBitBinary",
    classification_binary="lc_ClassificationBitBinary",
    sub_completeness="lc_SubCompletenessScore",
    completeness="lc_CumulativeCompletenessScore",
    inplace=False,
):
    """
    Adds the following completness score flags:
    - `SubCompletenessScore`: The percentage of valid landcover classifications and photos that are filled out.
    - `CumulativeCompletenessScore`: The percentage of non null values out of all the columns.

    Parameters
    ----------
    df : pd.DataFrame
        A landcover DataFrame with the [`PhotoBitBinary`](#photo_bit_flags) and [`ClassificationBitBinary`](#classification_bit_flags) flags.
    photo_bit_binary : str, default="lc_PhotoBitBinary"
        The name of the column that stores the PhotoBitBinary flag.
    classification_binary : str, default="lc_PhotoBitBinary"
        The name of the column that stores the ClassificationBitBinary flag.
    sub_completeness : str, default="lc_PhotoBitBinary"
        The name of the column that will store the generated SubCompletenessScore flag.
    completeness : str, default="lc_PhotoBitBinary"
        The name of the column that will store the generated CompletenessScore flag.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the completeness score flags. If `inplace=True` it returns None.
    """

    def sum_bit_mask(bit_mask="0"):
        sum = 0.0
        for char in bit_mask:
            sum += int(char)
        return sum

    if not inplace:
        df = df.copy()

    scores = {}
    scores["sub_score"] = []
    # Cummulative Completion Score
    scores["cumulative_score"] = round(df.count(1) / len(df.columns), 2)
    # Sub-Score
    for index in df.index:
        bit_mask = df[photo_bit_binary][index] + df[classification_binary][index]
        sub_score = round(sum_bit_mask(bit_mask=bit_mask), 2)
        sub_score /= len(bit_mask)
        scores["sub_score"].append(sub_score)

    df[sub_completeness], df[completeness] = (
        scores["sub_score"],
        scores["cumulative_score"],
    )

    if not inplace:
        return df


def apply_cleanup(lc_df, unpack=True):
    """Applies a full cleanup procedure to the landcover data.
    It follows the following steps:
    - Removes Homogenous Columns
    - Renames Latitude and Longitudes
    - Cleans the Column Naming
    - Unpacks landcover classifications
    - Rounds Columns
    - Standardizes Null Values

    This returns a copy

    Parameters
    ----------
    lc_df : pd.DataFrame
        A DataFrame containing **raw** Landcover Data from the API.
    unpack : bool
        If True, the Landcover data will expand the classifications into separate columns (results in around 300 columns). If False, it will just unpack overall landcover.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the cleaned Landcover Data
    """
    lc_df = lc_df.copy()

    remove_homogenous_cols(lc_df, inplace=True)
    rename_latlon_cols(lc_df, inplace=True)
    cleanup_column_prefix(lc_df, inplace=True)
    lc_df, overall_cols, directional_cols = unpack_classifications(lc_df, unpack=unpack)

    round_cols(lc_df, inplace=True)
    standardize_null_vals(lc_df, inplace=True)
    return lc_df


def add_flags(lc_df):
    """Adds the following flags to the landcover data:
    - Photo Bit Flags
    - Classification Bit Flags
    - Completeness Score Flags

    Returns a copy of the DataFrame

    Parameters
    ----------
    lc_df : pd.DataFrame
        A DataFrame containing cleaned up Landcover Data ideally from the [apply_cleanup](#apply_cleanup) method.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the Land Cover flags.
    """
    lc_df = lc_df.copy()
    photo_bit_flags(lc_df, inplace=True)
    classification_bit_flags(lc_df, inplace=True)
    get_main_classifications(lc_df, inplace=True)
    completion_scores(lc_df, inplace=True)
    return lc_df


def direction_frequency(lc_df, direction_list, bit_binary, entry_type):
    """
    Plots the amount of a variable of interest for each direction.

    Parameters
    ----------
    lc_df : pd.DataFrame
        The DataFrame containing Land Cover Data.
    direction_list : list of str
        The column names of the different variables of interest for each direction.
    bit_binary: str
        The Bit Binary Flag associated with the variable of interest.
    entry_type: str
        The variable of interest (e.g. Photos or Classifications)
    """
    direction_photos = pd.DataFrame()
    direction_photos["category"] = direction_list
    direction_counts = [0 for i in range(len(direction_photos))]
    for mask in lc_df[bit_binary]:
        for i in range(len(mask) - 1, -1, -1):
            direction_counts[i] += int(mask[i])
    direction_counts
    direction_photos["count"] = [math.log10(value) for value in direction_counts]
    direction_photos

    plt.figure(figsize=(15, 6))
    title = f"Land Cover -- {entry_type} Direction Frequency (Log Scale)"
    plt.title(title)
    plt.ylabel("Count (Log Scale)")
    sns.barplot(data=direction_photos, x="category", y="count", color="lightblue")


def diagnostic_plots(
    lc_df,
    up_url="lc_UpwardPhotoUrl",
    down_url="lc_DownwardPhotoUrl",
    north_url="lc_NorthPhotoUrl",
    south_url="lc_SouthPhotoUrl",
    east_url="lc_EastPhotoUrl",
    west_url="lc_WestPhotoUrl",
    photo_bit="lc_PhotoBitBinary",
    north_classification="lc_NorthClassifications",
    south_classification="lc_SouthClassifications",
    east_classification="lc_EastClassifications",
    west_classification="lc_WestClassifications",
    classification_bit="lc_ClassificationBitBinary",
):
    """
    Generates (but doesn't display) diagnostic plots to gain insight into the current data.

    Plots:
    - Valid Photo Count Distribution
    - Photo Distribution by direction
    - Classification Distribution by direction
    - Photo Status Distribution
    - Completeness Score Distribution
    - Subcompleteness Score Distribution

    Parameters
    ----------
    lc_df : pd.DataFrame
        The DataFrame containing Flagged and Cleaned Land Cover Data.
    """
    plot_freq_bar(
        lc_df, "Land Cover", "lc_PhotoCount", "Valid Photo Count", log_scale=True
    )
    direction_frequency(
        lc_df,
        [up_url, down_url, north_url, south_url, east_url, west_url],
        photo_bit,
        "Photo",
    )
    direction_frequency(
        lc_df,
        [
            north_classification,
            south_classification,
            east_classification,
            west_classification,
        ],
        classification_bit,
        "Classification",
    )
    multiple_bar_graph(
        lc_df,
        "Land Cover",
        ["lc_PhotoCount", "lc_RejectedCount", "lc_EmptyCount"],
        "Photo Summary",
        log_scale=True,
    )

    completeness_histogram(
        lc_df, "Land Cover", "lc_CumulativeCompletenessScore", "Cumulative Completeness"
    )
    completeness_histogram(
        lc_df, "Land Cover", "lc_SubCompletenessScore", "Sub Completeness"
    )


def qa_filter(
    lc_df,
    has_classification=False,
    has_photo=False,
    has_all_photos=False,
    has_all_classifications=False,
):
    """
    Can filter a cleaned and flagged mosquito habitat mapper DataFrame based on the following criteria:
    - `Has Classification`: If the entry has atleast one direction classified
    - `Has Photo` : If the entry has atleast one photo taken
    - `Has All Photos` : If the entry has all photos taken (up, down, north, south, east, west)
    - `Has All Classifications` : If the entry has all directions classified

    Returns a copy of the DataFrame

    Parameters
    ----------
    has_classification : bool, default=False
        If True, only entries with atleast one classification will be included.
    has_photo : bool, default=False
        If True, only entries with atleast one photo will be included.
    has_all_photos : bool, default=False
        If True, only entries with all photos will be included.
    has_all_classifications : bool, default=False
        If True, only entries with all classifications will be included.

    Returns
    -------
    pd.DataFrame
        A DataFrame of the applied filters.
    """

    if has_classification and not has_all_classifications:
        lc_df = lc_df[lc_df["lc_ClassificationBitDecimal"] > 0]
    elif has_all_classifications:
        lc_df = lc_df[lc_df["lc_ClassificationBitDecimal"] == 15]
    if has_photo and not has_all_photos:
        lc_df = lc_df[lc_df["lc_PhotoBitDecimal"] > 0]
    elif has_all_photos:
        lc_df = lc_df[lc_df["lc_PhotoBitDecimal"] == 63]

    return lc_df


def _accumulate_ties(classification_list):
    classifications = list()
    i = 0
    while i < len(classification_list) - 1:
        if classification_list[i][1] == classification_list[i + 1][1]:
            classifications.append(classification_list[i][0])
            classifications.append(classification_list[i + 1][0])
            i += 1
        else:
            break

    output = ", ".join([classification for classification in classifications])
    if not output:
        if len(classification_list) != 0:
            output = classification_list[0][0]
        else:
            output = "NA"
    # TODO replace w regex methods
    return output, i + 1


def _rank_direction(classification_dict, direction_classifications):
    if pd.isna(direction_classifications):
        return "NA", "NA"
    classifications_list = []
    classifications = direction_classifications.split(";")
    for classification_data in classifications:
        percent = extract_classification_percentage(classification_data)
        classification = extract_classification_name(classification_data)
        if classification in classification_dict:
            classification_dict[classification] += percent
        else:
            classification_dict[classification] = percent
        classifications_list.append((classification, percent))
    classifications_list = sorted(
        classifications_list, key=lambda x: x[1], reverse=True
    )
    if len(classifications_list) < 2:
        return classifications_list[0][0], "NA"

    primary_classification, i = _accumulate_ties(classifications_list)
    secondary_classification, temp = _accumulate_ties(classifications_list[i:])

    return primary_classification, secondary_classification


def _rank_classifications(*args):
    classification_dict = {}
    rank_directions = [
        classification
        for arg in args
        for classification in _rank_direction(classification_dict, arg)
    ]
    primary, secondary = ("NA", 0), ("NA", 0)
    if classification_dict:
        if len(classification_dict) < 2:
            primary = (
                list(classification_dict.keys())[0],
                list(classification_dict.values())[0],
            )
        else:
            sorted_classifications = sorted(
                classification_dict.items(), key=lambda x: x[1], reverse=True
            )
            primary, i = _accumulate_ties(sorted_classifications)
            primary = primary, sorted_classifications[0][1]
            if i < len(sorted_classifications):
                secondary, temp = _accumulate_ties(sorted_classifications[i:])
                secondary = secondary, sorted_classifications[i][1]
    return (
        *rank_directions,
        primary[0],
        secondary[0],
        primary[1] / len(args),
        secondary[1] / len(args),
    )


def get_main_classifications(
    lc_df,
    north_classification="lc_NorthClassifications",
    east_classification="lc_EastClassifications",
    south_classification="lc_SouthClassifications",
    west_classification="lc_WestClassifications",
    north_primary="lc_NorthPrimary",
    north_secondary="lc_NorthSecondary",
    east_primary="lc_EastPrimary",
    east_secondary="lc_EastSecondary",
    south_primary="lc_SouthPrimary",
    south_secondary="lc_SouthSecondary",
    west_primary="lc_WestPrimary",
    west_secondary="lc_WestSecondary",
    primary_classification="lc_PrimaryClassification",
    secondary_classification="lc_SecondaryClassification",
    primary_percentage="lc_PrimaryPercentage",
    secondary_percentage="lc_SecondaryPercentage",
    inplace=False,
):
    if not inplace:
        lc_df = lc_df.copy()
    vectorized_rank = np.vectorize(_rank_classifications)
    (
        lc_df[north_primary],
        lc_df[north_secondary],
        lc_df[east_primary],
        lc_df[east_secondary],
        lc_df[south_primary],
        lc_df[south_secondary],
        lc_df[west_primary],
        lc_df[west_secondary],
        lc_df[primary_classification],
        lc_df[secondary_classification],
        lc_df[primary_percentage],
        lc_df[secondary_percentage],
    ) = vectorized_rank(
        lc_df[north_classification].to_numpy(),
        lc_df[east_classification].to_numpy(),
        lc_df[south_classification].to_numpy(),
        lc_df[west_classification].to_numpy(),
    )

    if not inplace:
        return lc_df



import logging

import pandas as pd
import requests

import go_utils.lc as lc
import go_utils.mhm as mhm
from go_utils.constants import (
    end_date,
    landcover_protocol,
    mosquito_protocol,
    start_date,
)


def parse_api_data(response_json):
    try:
        results = response_json["results"]
        df = pd.DataFrame(results)
    except KeyError:
        raise RuntimeError("Data Download Failed. The GLOBE API is most likely down.")

    # Expand the 'data' column by listing the contents and passing as a new dataframe
    df = pd.concat([df, pd.DataFrame(list(df["data"]))], axis=1)
    # Drop the previously nested data column
    df = df.drop(labels="data", axis=1)

    # Display the dataframe
    return df


def is_valid_latlon_box(latlon_box):

    valid_lat_checks = (
        latlon_box["min_lat"] < latlon_box["max_lat"]
        and latlon_box["max_lat"] <= 90
        and latlon_box["min_lat"] >= -90
    )
    valid_lon_checks = (
        latlon_box["min_lon"] < latlon_box["max_lon"]
        and latlon_box["max_lon"] <= 180
        and latlon_box["min_lon"] >= -180
    )

    return valid_lon_checks and valid_lat_checks


def get_api_data(
    protocol,
    start_date=start_date,
    end_date=end_date,
    is_clean=True,
    latlon_box={"min_lat": -90, "max_lat": 90, "min_lon": -180, "max_lon": 180},
):
    """Utility function for interfacing with the GLOBE API.
    More information about the API can be viewed [here](https://www.globe.gov/es/globe-data/globe-api).

    Parameters
    ----------
    protocol : str
               The desired GLOBE Observer Protocol. Protocols for the App protocols include: `land_covers` (Landcover), `mosquito_habitat_mapper` (Mosquito Habitat Mapper), `sky_conditions` (Clouds), `tree_heights` (Trees).
    start_date : str, default= 2017-05-31
                 The desired start date of the dataset in the format of (YYYY-MM-DD).
    end_date : str, default= today's date in YYYY-MM-DD form.
               The desired end date of the dataset in the format of (YYYY-MM-DD).
    latlon_box : dict of {str, double}, optional
                 The longitudes and latitudes of a bounding box for the dataset. The minimum/maximum latitudes and longitudes must be specified with the following keys: "min_lat", "min_lon", "max_lat", "max_lon". The default value specifies all latitude and longitude coordinates.

    Returns
    -------
    pd.DataFrame
      A DataFrame containing Raw GLOBE Observer Data of the specified parameters
    """

    if is_valid_latlon_box(latlon_box):
        url = f"https://api.globe.gov/search/v1/measurement/protocol/measureddate/lat/lon/?protocols={protocol}&startdate={start_date}&enddate={end_date}&minlat={str(latlon_box['min_lat'])}&maxlat={str(latlon_box['max_lat'])}&minlon={str(latlon_box['min_lon'])}&maxlon={str(latlon_box['max_lon'])}&geojson=FALSE&sample=FALSE"
    else:
        logging.warning(
            "You did not enter any valid/specific coordinates, so we gave you all the observations for your protocol, date_range, and any countryNames you may have specified.\n"
        )
        url = f"https://api.globe.gov/search/v1/measurement/protocol/measureddate/?protocols={protocol}&startdate={start_date}&enddate={end_date}&geojson=FALSE&sample=FALSE"

    # Downloads data from the GLOBE API
    response = requests.get(url)

    if not response:
        raise RuntimeError(
            "Failed to get data from the API. Double check your specified settings to make sure they are valid."
        )

    # Convert measured date data into datetime
    df = parse_api_data(response.json())
    convert_dates_to_datetime(df)

    if is_clean:
        df = default_data_clean(df, protocol)
    return df


def convert_dates_to_datetime(df):
    date_columns = [col for col in df.columns if "Date" in col or "MeasuredAt" in col]
    for column in date_columns:
        df[column] = pd.to_datetime(df[column], errors="coerce")


def default_data_clean(df, protocol):
    module_mapper = {mosquito_protocol: mhm, landcover_protocol: lc}
    if protocol in module_mapper:
        df = module_mapper[protocol].apply_cleanup(df)
        df = module_mapper[protocol].add_flags(df)
    else:
        logging.warning("The protocol you entered is not supported for cleanup.")

    return df



import math
import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from go_utils.cleanup import (
    rename_latlon_cols,
    replace_column_prefix,
    round_cols,
    standardize_null_vals,
)
from go_utils.plot import completeness_histogram, plot_freq_bar, plot_int_distribution

__doc__ = r"""

## Mosquito Specific Cleanup Procedures

### Converting Larvae Data to Integers
Larvae Data is stored as a string in the raw GLOBE Observer dataset. To facillitate analysis, [this method](#larvae_to_num) converts this data to numerical data.

It needs to account for 4 types of data:
1. Regular Data: Converts it to a number
2. Extraneously large data ($\geq 100$ as its hard to count more than that amount accurately): To maintain the information from that entry, the `LarvaeCountMagnitude` flag is used to indicate the real value
3. Ranges (e.g. "25-50"): Chooses the lower bound and set the `LarvaeCountIsRangeFlag` to true.
4. Null Values: Sets null values to $-9999$


It generates the following flags:
- `LarvaeCountMagnitude`: The integer flag contains the order of magnitude (0-4) by which the larvae count exceeds the maximum Larvae Count of 100. This is calculated by $1 + \lfloor \log{\frac{num}{100}} \rfloor$. As a result:
    - `0`: Corresponds to a Larvae Count $\leq 100$
    - `1`: Corresponds to a Larvae Count between $100$ and $999$
    - `2`: Corresponds to a Larvae Count between $1000$ and $9999$
    - `3`: Corresponds to a Larvae Count between $10,000$ and $99,999$
    - `4`: Corresponds to a Larvae Count $\geq 100,000$
- `LarvaeCountIsRange`: Either a $1$ which indicates the entry was a range (e.g. 25-50) or $0$ which indicates the entry wasn't a range.

Additionally, there were extremely large values that Python was unable to process (`1e+27`) and so there was an initial preprocessing step to set those numbers to 100000 (which corresponds to the maximum magnitude flag).
"""


def cleanup_column_prefix(df, inplace=False):
    """Method for shortening raw mosquito habitat mapper column names.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing raw mosquito habitat mapper data.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame or None
        A DataFrame with the cleaned up column prefixes. If `inplace=True` it returns None.
    """

    return replace_column_prefix(df, "mosquitohabitatmapper", "mhm", inplace=inplace)


def _entry_to_num(entry):
    try:
        if entry == "more than 100":
            return 101, 1, 1
        if pd.isna(entry):
            return -9999, 0, 0
        elif float(entry) > 100:
            return 101, min(math.floor(math.log10(float(entry) / 100)) + 1, 4), 0
        return float(entry), 0, 0
    except ValueError:
        return float(re.sub(r"-.*", "", entry)), 0, 1


def larvae_to_num(
    mhm_df,
    larvae_count_col="mhm_LarvaeCount",
    magnitude="mhm_LarvaeCountMagnitude",
    range_flag="mhm_LarvaeCountIsRangeFlag",
    inplace=False,
):
    """Converts the Larvae Count of the Mosquito Habitat Mapper Dataset from being stored as a string to integers.

    See [here](#converting-larvae-data-to-integers) for more information.

    Parameters
    ----------
    mhm_df : pd.DataFrame
        A DataFrame of Mosquito Habitat Mapper data that needs the larvae counts to be set to numbers
    larvae_count_col : str, default="mhm_LarvaeCount"
        The name of the column storing the larvae count. **Note**: The columns will be output in the format: `prefix_ColumnName` where `prefix` is all the characters that preceed the words `LarvaeCount` in the specified name.
    magnitude: str, default="mhm_LarvaeCountMagnitude"
        The name of the column which will store the generated LarvaeCountMagnitude output
    range_flag : str, default="mhm_LarvaeCountIsRangeFlag"
        The name of the column which will store the generated LarvaeCountIsRange flag
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the larvae count as integers. If `inplace=True` it returns None.
    """

    if not inplace:
        mhm_df = mhm_df.copy()
    # Preprocessing step to remove extremely erroneous values
    for i in mhm_df.index:
        count = mhm_df[larvae_count_col][i]
        if not pd.isna(count) and type(count) is str and "e+" in count:
            mhm_df.at[i, larvae_count_col] = "100000"

    larvae_conversion = np.vectorize(_entry_to_num)
    (
        mhm_df[larvae_count_col],
        mhm_df[magnitude],
        mhm_df[range_flag],
    ) = larvae_conversion(mhm_df[larvae_count_col].to_numpy())

    if not inplace:
        return mhm_df


def has_genus_flag(df, genus_col="mhm_Genus", bit_col="mhm_HasGenus", inplace=False):
    """
    Creates a bit flag: `mhm_HasGenus` where 1 denotes a recorded Genus and 0 denotes the contrary.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame
    genus_col : str, default="mhm_Genus"
        The name of the column in the mosquito habitat mapper DataFrame that contains the genus records.
    bit_col : str, default="mhm_HasGenus"
        The name of the column which will store the generated HasGenus flag
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the HasGenus flag. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()
    df[bit_col] = (~pd.isna(df[genus_col].to_numpy())).astype(int)

    if not inplace:
        return df


def infectious_genus_flag(
    df, genus_col="mhm_Genus", bit_col="mhm_IsGenusOfInterest", inplace=False
):
    """
    Creates a bit flag: `mhm_IsGenusOfInterest` where 1 denotes a Genus of a infectious mosquito and 0 denotes the contrary.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame
    genus_col : str, default="mhm_Genus"
        The name of the column in the mosquito habitat mapper DataFrame that contains the genus records.
    bit_col : str, default="mhm_HasGenus"
        The name of the column which will store the generated IsGenusOfInterest flag
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the IsGenusOfInterest flag. If `inplace=True` it returns None.
    """
    if not inplace:
        df = df.copy()
    infectious_genus_flag = np.vectorize(
        lambda genus: genus in ["Aedes", "Anopheles", "Culex"]
    )
    df[bit_col] = infectious_genus_flag(df[genus_col].to_numpy()).astype(int)

    if not inplace:
        return df


def is_container_flag(
    df,
    watersource_col="mhm_WaterSourceType",
    bit_col="mhm_IsWaterSourceContainer",
    inplace=False,
):
    """
    Creates a bit flag: `mhm_IsWaterSourceContainer` where 1 denotes if a watersource is a container (e.g. ovitrap, pots, tires, etc.) and 0 denotes the contrary.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame
    watersource_col : str, default="mhm_WaterSourceType"
        The name of the column in the mosquito habitat mapper DataFrame that contains the watersource type records.
    bit_col : str, default="mhm_IsWaterSourceContainer"
        The name of the column which will store the generated IsWaterSourceContainer flag
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the IsContainer flag. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()

    mark_containers = np.vectorize(
        lambda container: not pd.isna(container) and "container" in container
    )
    df[bit_col] = mark_containers(df[watersource_col].to_numpy()).astype(int)

    if not inplace:
        return df


def has_watersource_flag(
    df, watersource_col="mhm_WaterSource", bit_col="mhm_HasWaterSource", inplace=False
):
    """
    Creates a bit flag: `mhm_HasWaterSource` where 1 denotes if there is a watersource and 0 denotes the contrary.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame
    watersource_col : str, default="mhm_WaterSource"
        The name of the column in the mosquito habitat mapper DataFrame that contains the watersource records.
    bit_col : str, default="mhm_IsWaterSourceContainer"
        The name of the column which will store the generated HasWaterSource flag
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the HasWaterSource flag. If `inplace=True` it returns None.
    """

    if not inplace:
        df = df.copy()
    has_watersource = np.vectorize(lambda watersource: int(not pd.isna(watersource)))
    df[bit_col] = has_watersource(df[watersource_col].to_numpy())

    if not inplace:
        return df


def photo_bit_flags(
    df,
    watersource_photos="mhm_WaterSourcePhotoUrls",
    larvae_photos="mhm_LarvaFullBodyPhotoUrls",
    abdomen_photos="mhm_AbdomenCloseupPhotoUrls",
    photo_count="mhm_PhotoCount",
    rejected_count="mhm_RejectedCount",
    pending_count="mhm_PendingCount",
    photo_bit_binary="mhm_PhotoBitBinary",
    photo_bit_decimal="mhm_PhotoBitDecimal",
    inplace=False,
):
    """
    Creates the following flags:
    - `PhotoCount`: The number of valid photos per record.
    - `RejectedCount`: The number of photos that were rejected per record.
    - `PendingCount`: The number of photos that are pending approval per record.
    - `PhotoBitBinary`: A string that represents the presence of a photo in the order of watersource, larvae, and abdomen. For example, if the entry is `110`, that indicates that there is a water source photo and a larvae photo, but no abdomen photo.
    - `PhotoBitDecimal`: The numerical representation of the mhm_PhotoBitBinary string.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame
    watersource_photos : str, default="mhm_WaterSourcePhotoUrls"
        The name of the column in the mosquito habitat mapper DataFrame that contains the watersource photo url records.
    larvae_photos : str, default="mhm_LarvaFullBodyPhotoUrls"
        The name of the column in the mosquito habitat mapper DataFrame that contains the larvae photo url records.
    abdomen_photos : str, default="mhm_AbdomenCloseupPhotoUrls"
        The name of the column in the mosquito habitat mapper DataFrame that contains the abdomen photo url records.
    photo_count : str, default="mhm_PhotoCount"
        The name of the column that will store the PhotoCount flag.
    rejected_count : str, default="mhm_RejectedCount"
        The name of the column that will store the RejectedCount flag.
    pending_count : str, default="mhm_PendingCount"
        The name of the column that will store the PendingCount flag.
    photo_bit_binary : str, default="mhm_PhotoBitBinary"
        The name of the column that will store the PhotoBitBinary flag.
    photo_bit_decimal : str, default="mhm_PhotoBitDecimal"
        The name of the column that will store the PhotoBitDecimal flag.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with the photo flags. If `inplace=True` it returns None.
    """

    def pic_data(*args):
        pic_count = 0
        rejected_count = 0
        pending_count = 0
        valid_photo_bit_mask = ""

        # bit_power = len(args) - 1
        # For url string -- if we see ANY http, add 1
        # also count all valid photos, rejected photos,
        # If there are NO http then add 0, to empty photo field
        for url_string in args:
            if not pd.isna(url_string):
                if "http" not in url_string:
                    valid_photo_bit_mask += "0"
                else:
                    valid_photo_bit_mask += "1"

                pic_count += url_string.count("http")
                pending_count += url_string.count("pending")
                rejected_count += url_string.count("rejected")
            else:
                valid_photo_bit_mask += "0"

        return (
            pic_count,
            rejected_count,
            pending_count,
            valid_photo_bit_mask,
            int(valid_photo_bit_mask, 2),
        )

    if not inplace:
        df = df.copy()

    get_photo_data = np.vectorize(pic_data)
    (
        df[photo_count],
        df[rejected_count],
        df[pending_count],
        df[photo_bit_binary],
        df[photo_bit_decimal],
    ) = get_photo_data(
        df[watersource_photos].to_numpy(),
        df[larvae_photos].to_numpy(),
        df[abdomen_photos].to_numpy(),
    )

    if not inplace:
        return df


def completion_score_flag(
    df,
    photo_bit_binary="mhm_PhotoBitBinary",
    has_genus="mhm_HasGenus",
    sub_completeness="mhm_SubCompletenessScore",
    completeness="mhm_CumulativeCompletenessScore",
    inplace=False,
):
    """
    Adds the following completness score flags:
    - `SubCompletenessScore`: The percentage of the watersource photos, larvae photos, abdomen photos, and genus columns that are filled out.
    - `CumulativeCompletenessScore`: The percentage of non null values out of all the columns.

    Parameters
    ----------
    df : pd.DataFrame
        A mosquito habitat mapper DataFrame with the [`PhotoBitDecimal`](#photo_bit_flags) and [`HasGenus`](#has_genus_flags) flags.
    photo_bit_binary: str, default="mhm_PhotoBitBinary"
        The name of the column in the mosquito habitat mapper DataFrame that contains the PhotoBitBinary flag.
    sub_completeness : str, default="mhm_HasGenus"
        The name of the column in the mosquito habitat mapper DataFrame that will contain the generated SubCompletenessScore flag.
    completeness : str, default="mhm_SubCompletenessScore"
        The name of the column in the mosquito habitat mapper DataFrame that will contain the generated CumulativeCompletenessScore flag.
    inplace : bool, default=False
        Whether to return a new DataFrame. If True then no DataFrame copy is not returned and the operation is performed in place.

    Returns
    -------
    pd.DataFrame
        A DataFrame with completion score flags. If `inplace=True` it returns None.
    """

    def sum_bit_mask(bit_mask="0"):
        total = 0.0
        for char in bit_mask:
            total += int(char)
        return total

    if not inplace:
        df = df.copy()

    scores = {}
    scores["sub_score"] = []
    # Cummulative Completion Score
    scores["cumulative_score"] = round(df.count(axis=1) / len(df.columns), 2)
    # Sub-Score
    for index in df.index:
        bit_mask = df[photo_bit_binary][index]
        sub_score = df[has_genus][index] + sum_bit_mask(bit_mask=bit_mask)
        sub_score /= 4.0
        scores["sub_score"].append(sub_score)

    df[sub_completeness], df[completeness] = (
        scores["sub_score"],
        scores["cumulative_score"],
    )

    if not inplace:
        return df


def apply_cleanup(mhm_df):
    """Applies a full cleanup procedure to the mosquito habitat mapper data. Only returns a copy.
    It follows the following steps:
    - Removes Homogenous Columns
    - Renames Latitude and Longitudes
    - Cleans the Column Naming
    - Converts Larvae Count to Numbers
    - Rounds Columns
    - Standardizes Null Values

    Parameters
    ----------
    mhm_df : pd.DataFrame
        A DataFrame containing **raw** Mosquito Habitat Mapper Data from the API.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the cleaned up Mosquito Habitat Mapper Data
    """
    mhm_df = mhm_df.copy()

    rename_latlon_cols(mhm_df, inplace=True)
    cleanup_column_prefix(mhm_df, inplace=True)
    larvae_to_num(mhm_df, inplace=True)
    round_cols(mhm_df, inplace=True)
    standardize_null_vals(mhm_df, inplace=True)
    return mhm_df


def add_flags(mhm_df):
    """Adds the following flags to the Mosquito Habitat Mapper Data:
    - Has Genus
    - Is Infectious Genus/Genus of Interest
    - Is Container
    - Has WaterSource
    - Photo Bit Flags
    - Completion Score Flag

    This returns a copy of the original DataFrame with the flags added onto it.

    Parameters
    ----------
    mhm_df : pd.DataFrame
        A DataFrame containing cleaned up Mosquito Habitat Mapper Data ideally from the method.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the flagged Mosquito Habitat Mapper Data
    """
    mhm_df = mhm_df.copy()
    has_genus_flag(mhm_df, inplace=True)
    infectious_genus_flag(mhm_df, inplace=True)
    is_container_flag(mhm_df, inplace=True)
    has_watersource_flag(mhm_df, inplace=True)
    photo_bit_flags(mhm_df, inplace=True)
    completion_score_flag(mhm_df, inplace=True)
    return mhm_df


def plot_valid_entries(df, bit_col, entry_type):
    """
    Plots the number of entries with photos and the number of entries without photos

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing Mosquito Habitat Mapper Data with the PhotoBitDecimal Flag.
    """
    plt.figure()
    num_valid = len(df[df[bit_col] > 0])
    plt.title(f"Entries with {entry_type} vs No {entry_type}")
    plt.ylabel("Number of Entries")
    plt.bar(entry_type, num_valid, color="#e34a33")
    plt.bar(f"No {entry_type}", len(df) - num_valid, color="#fdcc8a")


def photo_subjects(mhm_df):
    """
    Plots the amount of photos for each photo area (Larvae, Abdomen, Watersource)

    Parameters
    ----------
    mhm_df : pd.DataFrame
        The DataFrame containing Mosquito Habitat Mapper Data with the PhotoBitDecimal Flag.
    """

    total_dict = {"Larvae Photos": 0, "Abdomen Photos": 0, "Watersource Photos": 0}

    for number in mhm_df["mhm_PhotoBitDecimal"]:
        total_dict["Watersource Photos"] += number & 4
        total_dict["Larvae Photos"] += number & 2
        total_dict["Abdomen Photos"] += number & 1

    for key in total_dict.keys():
        if total_dict[key] != 0:
            total_dict[key] = math.log10(total_dict[key])
        else:
            total_dict[key] = 0
    plt.figure(figsize=(10, 5))
    plt.title("Mosquito Habitat Mapper - Photo Subject Frequencies (Log Scale)")
    plt.xlabel("Photo Type")
    plt.ylabel("Frequency (Log Scale)")
    plt.bar(total_dict.keys(), total_dict.values(), color="lightblue")


def diagnostic_plots(mhm_df):
    """
    Generates (but doesn't display) diagnostic plots to gain insight into the current data.

    Plots:
    - Larvae Count Distribution (where a negative entry denotes null data)
    - Photo Subject Distribution
    - Number of valid photos vs no photos
    - Completeness Score Distribution
    - Subcompleteness Score Distribution

    Parameters
    ----------
    mhm_df : pd.DataFrame
        The DataFrame containing Flagged and Cleaned Mosquito Habitat Mapper Data.
    """
    plot_int_distribution(mhm_df, "mhm_LarvaeCount", "Larvae Count")
    photo_subjects(mhm_df)
    plot_freq_bar(mhm_df, "Mosquito Habitat Mapper", "mhm_Genus", "Genus Types")
    plot_valid_entries(mhm_df, "mhm_HasGenus", "Genus Classifications")
    plot_valid_entries(mhm_df, "mhm_PhotoBitDecimal", "Valid Photos")
    completeness_histogram(
        mhm_df,
        "Mosquito Habitat Mapper",
        "mhm_CumulativeCompletenessScore",
        "Cumulative Completeness",
    )
    completeness_histogram(
        mhm_df,
        "Mosquito Habitat Mapper",
        "mhm_SubCompletenessScore",
        "Sub Completeness",
    )


def qa_filter(
    mhm_df,
    has_genus=False,
    min_larvae_count=-9999,
    has_photos=False,
    is_container=False,
):
    """
    Can filter a cleaned and flagged mosquito habitat mapper DataFrame based on the following criteria:
    - `Has Genus`: If the entry has an identified genus
    - `Min Larvae Count` : Minimum larvae count needed for an entry
    - `Has Photos` : If the entry contains valid photo entries
    - `Is Container` : If the entry's watersource was a container

    Returns a copy of the DataFrame

    Parameters
    ----------
    has_genus : bool, default=False
        If True, only entries with an identified genus will be returned.
    min_larvae_count : int, default=-9999
        Only entries with a larvae count greater than or equal to this parameter will be included.
    has_photos : bool, default=False
        If True, only entries with recorded photos will be returned
    is_container : bool, default=False
        If True, only entries with containers will be returned

    Returns
    -------
    pd.DataFrame
        A DataFrame of the applied filters.
    """

    mhm_df = mhm_df[mhm_df["mhm_LarvaeCount"] >= min_larvae_count]

    if has_genus:
        mhm_df = mhm_df[mhm_df["mhm_HasGenus"] == 1]
    if has_photos:
        mhm_df = mhm_df[mhm_df["mhm_PhotoBitDecimal"] > 0]
    if is_container:
        mhm_df = mhm_df[mhm_df["mhm_IsWaterSourceContainer"] == 1]

    return mhm_df



from datetime import datetime

# Default Start and End Dates for GO Data
start_date = "2017-05-31"
end_date = datetime.now().strftime("%Y-%m-%d")

mosquito_protocol = "mosquito_habitat_mapper"
landcover_protocol = "land_covers"

region_dict = {
    "Africa": [
        "Benin",
        "Botswana",
        "Burkina Faso",
        "Cameroon",
        "Cape Verde",
        "Chad",
        "Congo DRC",
        "Ethiopia",
        "Gabon",
        "Gambia",
        "Ghana",
        "Guinea",
        "Kenya",
        "Liberia",
        "Madagascar",
        "Mali",
        "Mauritius",
        "Namibia",
        "Niger",
        "Nigeria",
        "Rwada",
        "Senegal",
        "Seychelles",
        "South Africa",
        "Tanzania",
        "Togo",
        "Uganda",
    ],
    "Asia and the Pacific": [
        "Australia",
        "Bangladesh",
        "Fiji",
        "India",
        "Japan",
        "Maldives",
        "Marshall Islands",
        "Micronesia",
        "Mongolia",
        "Nepal",
        "New Zealand",
        "Palau",
        "Philippines",
        "Republic of Korea",
        "Sri Lanka",
        "Taiwan Partnership",
        "Thailand",
        "Vietnam",
    ],
    "Latin America and Caribbean": [
        "Argentina",
        "Bahamas",
        "Burmuda",
        "Bolivia",
        "Brazil",
        "Chile",
        "Colombia",
        "Costa Rica",
        "Dominican Republic",
        "Ecuador",
        "El Salvador",
        "Guatemala",
        "Honduras",
        "Mexico",
        "Panama",
        "Paraguay",
        "Peru",
        "Suriname",
        "Trinidad and Tobago",
        "Uruguay",
    ],
    "Europe and Eurasia": [
        "Austria",
        "Belgium",
        "Bulgaria",
        "Croatia",
        "Cyprus",
        "Czech Republic",
        "Denmark",
        "Estonia",
        "Finland",
        "France",
        "Georgia",
        "Germany",
        "Greece",
        "Hungary",
        "Iceland",
        "Ireland",
        "Israel",
        "Italy",
        "Kazakhstan",
        "Kyrgyz Republic",
        "Latvia",
        "Liechtenstein",
        "Lithuania",
        "Luxembourg",
        "Malta",
        "Moldovia",
        "Montenegro",
        "Netherlands",
        "North Macedonia",
        "Norway",
        "Poland",
        "Portugal",
        "Romania",
        "Russia",
        "Serbia",
        "Slovak Republic",
        "Spain",
        "Sweden",
        "Switzerland",
        "Turkey",
        "Ukraine",
        "United Kingdom",
    ],
    "Near East and North Africa": [
        "Bahrain",
        "Egypt",
        "Jordan",
        "Kuwait",
        "Lebanon",
        "Mauritania",
        "Morocco",
        "Oman",
        "Pakistan",
        "Qatar",
        "Saudi Arabia",
        "Tunisia",
        "United Arab Emirates",
    ],
    "North America": ["Canada", "United States"],
}

abbreviation_dict = {
    mosquito_protocol: "mhm",
    landcover_protocol: "lc",
}

__doc__ = """
Useful stored values for GLOBE Observer scripts

Stored Values
-------------
- mosquito_protocol: the string used in the GLOBE API for the Mosquito Habitat Mapper protocol
- landcover_protocol: the string used in the GLOBE API for the Land Cover Protocol.
- start_date: the default starting date of GLOBE Obsever data requests in YYYY-MM-DD Form.
- end_date: the current date in YYYY-MM-DD Form.
- regions_dict: contains all the different GLOBE regions and the countries associated with each region
"""



import os
import re
import subprocess
from os.path import abspath, dirname, exists, join

from setuptools import find_packages, setup

setup_dir = dirname(abspath(__file__))
git_dir = join(setup_dir, ".git")
version_file = join(setup_dir, "version.py")
package_name = "go_utils"


def get_git_tag():
    try:
        git_tag = str(
            subprocess.check_output(
                ["git", "describe", "--exact-match", "--abbrev=0"],
                stderr=subprocess.STDOUT,
            )
        ).strip("'b\\n")
    except subprocess.CalledProcessError as exc_info:
        match = re.search(
            "fatal: no tag exactly matches '(?P<commit>[a-z0-9]+)'",
            str(exc_info.output),
        )
        if match:
            raise Exception(
                "Bailing: there is no git tag for the current commit, {commit}".format(
                    commit=match.group("commit")
                )
            )
        raise Exception(str(exc_info.output))

    return git_tag


# Automatically generate a version.py based on the git version
if exists(git_dir):
    proc = subprocess.run(
        ["git", "describe", "--exact-match", "--abbrev=0"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    # If there is no main branch, the commit count defaults to 0
    if proc.returncode:
        version_str = "0"
    else:
        version_str = proc.stdout.decode("utf-8")

    # Version number: <year>.<# commits on main>
    version = version_str.strip()
elif "go_utils-" in os.getcwd():
    version = os.getcwd().split("go_utils-")[1].strip()
else:
    version = "0"
if version != "0":
    # Create the version.py file
    with open(version_file, "w+") as fp:
        fp.write(f'# Autogenerated by setup.py\n__version__ = "{version}"')

if exists(version_file):
    with open(version_file, "r") as fp:
        exec(fp.read(), globals())
else:
    __version__ = "main"

with open(join(setup_dir, "README.md"), "r") as readme_file:
    long_description = readme_file.read()

setup(
    name=package_name,
    version=__version__,
    description="Utilities for interfacing with GLOBE Observer Data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="IGES",
    maintainer="Matteo Kimura, Prachi Ingle",
    maintainer_email="mateus.sakata@gmail.com, prachi.ingle@utexas.edu",
    project_urls={
        "Source Code": "https://github.com/IGES-Geospatial/globe-observer-utils",
        "Documentation": "https://iges-geospatial.github.io/globe-observer-utils-docs/go_utils.html",
        "Bug Tracker": "https://github.com/IGES-Geospatial/globe-observer-utils/issues",
    },
    entry_points={
        "console_scripts": [
            "mhm-download=go_utils._cli:mhm_data_download",
            "lc-download=go_utils._cli:lc_data_download",
            "mhm-photo-download=go_utils._cli:mhm_photo_download",
            "lc-photo-download=go_utils._cli:lc_photo_download",
        ]
    },
    keywords="GlobeObserver GLOBE mosquito landcover",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    install_requires=[
        "numpy>=1.19.0",
        "pandas>=1.1.5",
        "requests>=2.23.0",
        "arcgis==1.9.1",
        "seaborn>=0.11.1",
        "pytz>=2021.3",
        "timezonefinder>=5.2.0",
    ],
    python_requires=">=3.6",
    license="MIT License",
    classifiers=[
        "Intended Audience :: Education",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: GIS",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
    ],
)



# flake8: noqa

from go_utils.download import get_api_data

__docformat__ = "restructuredtext"

__doc__ = """
go_utils contains utilities for interfacing with the GLOBE Observer datasets, primarily the Mosquito Habitat Mapper and Landcover.

Main Features
-------------
Here are the major utilities provided by the package:

- Easy downloading of the GLOBE Observer datasets as Pandas DataFrames
- Utilities for cleaning up Mosquito Habitat Mapper and Landcover Datasets
- CLI Scripts for downloading data and photos for the Mosquito Habitat Mapper and Landcover Protocols.

Example Notebooks
-----------------
[Example Notebooks](https://github.com/IGES-Geospatial/globe-observer-utils/tree/main/notebooks) have been made to show basic examples and applications of this library.

CLI Script Documentation
-----------

The CLI Scripts are intended to provide fast and accessible ways of collecting cleaned GLOBE Observer data from the commandline.

## Getting GLOBE Data CSVs
There are two commands that can be used to get data from Mosquito Habitat Mapper and Land Cover, respectively.
For Mosquito Habitat Mapper, there is `mhm-data-download` and for Land Cover, there is `lc-data-download`.

For each of these commands, there are several flags that can be specified to narrow down the data:
### General Flags

#### Output Path
You can use `--out` or `-o` followed by a file path to specify the output path of the download. If this is not specified, the script will generate diagnostic plots for you to get insight into the data that you would be downloading.

#### Start and End Dates
You can use `--start` or `-s` followed by a date in YYYY-MM-DD form to specify the start date. Similarly, you can use `--end` or `-e` to specify the end date.

#### Countries
You can use `--countries` or `-co` followed by a list of countries separated by commas to specify a download of specific country data. For example `-co "Brazil, Argentina"` would only download data from Brazil and Argentina.

#### Regions
You can use `--regions` or `-r` followed by a list of regions separated by commas to specify a download of specific GLOBE region data. For example `-r "North America"` would download data from North America. See the region dictionary to see the countries that comprise each GLOBE Region.

#### Bounding Box
You can use `--box` or `-b` followed by the coordinates of your bounding box in this format: `min latitude, min longitude, max latitude, max longitude`

### Mosquito Habitat Mapper Flags

#### Genus
You can use `--hasgenus` or `-hg` to only get data that has an identified genus.

#### Container
You can use `--iscontainer` or `-ic` to only get data which has a container as a watersource.

#### Photo
You can use `--hasphotos` or `-hp` to only get data which has recorded photo entries.

#### Minimum Larvae Count
You can use `--minlarvae` or `-ml` to only get data which has a certain amount of larvae.

#### Example
The command:
```py
mhm-download -s "2021-05-01" -e "2021-05-31" -hg -o MHM_Regular_Test.csv
```

Would download all mosquito mapper data with an identified genus during May 1st, 2021 to May 31st, 2021 into a `MHM_Regular_Test.csv` file.
### Land Cover Flags

#### Has a Classification
You can use `--hasclassification` or `-hc` to only get data which has atleast one Land Cover Classification.

#### Has Photo
You can use `--hasphoto` or `-hp` to only get data which has atleast one Photo entry.

#### Has All Classifications
You can use `--hasallclassifications` or `-hac` to only get data which has all classifications filled out.

#### Has All Photos
You can use `--hasallphotos` or `-hap` to only get data which has all photos filled out.

#### Example
The command:
```py
lc-download -s "2021-05-01" -e "2021-05-31" -hap -o LC_Regular_Test.csv
```

Would download all landcover data with fully filled out photo observations during May 1st, 2021 to May 31st, 2021 into a `LC_Regular_Test.csv` file.

## Downloading Photos
There are two commands that can be used to get photos for cleaned Mosquito Habitat Mapper and Land Cover CSV files, respectively.
For Mosquito Habitat Mapper, there is `mhm-photo-download` and for Land Cover, there is `lc-photo-download`.

### General Command
In order for the photo download scripts to work, you must specify an input path and output path.
For example, `mhm_photo_download "input csv path" "output directory"` will take the Mosquito Habitat Mapper data stored in the input CSV and download the photos to the output directory. If the output directory doesn't exist, it will be created.

#### All Flag
You can use the `--all` or `-a` to set all the following flags.

#### Custom Photo Naming Flags
* Use `--name_additional` or `-add` to add custom additional information to the photo names.
* Use `--nargs_include` or `-in` to include a field from the default photo name. Can include any of the following:
    * For **mosquito_habitat_mapper** -- Accepted Included Names include:
        * `url_type` -- Type of photo (e.g. Watersource, Larvae, Abdomen)
        * `watersource` -- Watersource for the observed mosquito habitat
        * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
        * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
        * `date_str` -- Date Range expressed as a String
        * `mhm_id` -- Unique ID for a MHM observation
        * `classification` -- Mosquito classification (or `"None"` if no classification available)
    * For **land_cover** -- Accepted Included Names include:
        * `direction` -- Direction where the photo was taken (e.g. North, South, East, West, Up, Down)
        * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
        * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
        * `date_str` -- Date Range expressed as a String
        * `lc_id` -- Unique ID for a LC observation

### Mosquito Specifics

#### Larvae Photos
You can use `--larvae` or `-l` to specify the download of larvae photos.

#### Watersource Photos
You can use `--watersource` or  `-w` to specify the download of watersource photos.

#### Abdomen Photos
You can use `--abdomen` or "-ab" to specify the download of abdomen photos.

#### Example
The command
```py
mhm-photo-download mhm_sample.csv mhm_outdir
```
Would download all mosquito mapper photos in the `mhm_sample.csv` file to a directory named `mhm_outdir`
### Landcover Specifics

#### Up Photos
You can use `--up` or `-u` to specify the download of upward photos.

#### Down Photos
You can use `--down` or `-d` to specify the download of downward photos.

#### North Photos
You can use `--north` or `-n` to specify the download of northward photos.

#### South Photos
You can use `--south` or `-s` to specify the download of southward photos.

#### East Photos
You can use `--east` or `-e` to specify the download of eastward photos.

#### West Photos
You can use `--west` or `-w` to specify the download of westward photos.

#### Example
The command:
```py
lc-photo-download lc_sample.csv lc_outdir -a
```
Would download all landcover photos in the `lc_sample.csv` file to a directory named `lc_outdir`
"""



import os
import re
import warnings

import numpy as np
import pandas as pd
import requests
from PIL import Image


def get_globe_photo_id(url: str):
    """
    Gets the GLOBE Photo ID from a url

    Parameters
    ----------
    url : str
      A url to a GLOBE Observer Image
    """
    if pd.isna(url):
        return None
    else:
        match_obj = re.search(r"(?<=\d\d\d\d\/\d\d\/\d\d\/).*(?=\/)", url)
        if match_obj:
            photo_id = match_obj.group(0)
            return photo_id
    return None


def remove_bad_characters(filename: str):
    """
    Removes erroneous characters from filenames. This includes the `/` character as this is assuming that the filename is being passed, not a path that may include that symbol as part of a directory.

    Parameters
    ----------
    filename : str
      A possible filename.

    Returns
    -------
    str
        The filename without any erroneous characters
    """
    if pd.isna(filename):
        return None
    return re.sub(r"[<>:?\"/\\|*]", "", filename)


def download_photo(url: str, directory: str, filename: str, resolution=None):
    """
    Downloads a photo to a directory.

    Parameters
    ----------
    url : str
        The URL to the photo
    directory : str
        The directory that the photo should be saved in
    filename : str
        The name of the photo
    resolution : tuple of int, default = None
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image. If the resolution is None, the original resolution of the photo is downloaded.
    """
    if any(pd.isna(x) for x in [url, directory, filename]):
        msg = f"Either url ({url}), directory ({directory}), or filename ({filename}) was None."
        warnings.warn(msg)
    else:
        downloaded_obj = requests.get(url, allow_redirects=True)
        filename = remove_bad_characters(filename)
        out_path = os.path.join(directory, filename)
        if not os.path.exists(directory):
            os.mkdir(directory)
        if pd.isna(resolution):
            with open(out_path, "wb") as file:
                file.write(downloaded_obj.content)
        else:
            get_img_at_resolution(url, out_path, resolution)


def get_img_at_resolution(url, path, resolution):
    """
    Downloads an image from a url at a specified resolution

    Parameters
    ----------
    url : str
        An image URL
    path : str
        The filepath to save the image to
    resolution : tuple of int
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image.
    """

    def get_img():
        with Image.open(requests.get(url, stream=True).raw) as img:
            img.resize(resolution).save(path)

    try:
        get_img()
    except Exception as e:  # Sometimes the image download fails and it has to be rerun
        warnings.warn(f"{url} failed due to {repr(e)}, retrying...")
        try:
            get_img()
            warnings.warn("retry successful")
        except Exception as e:
            warnings.warn(f"{url} failed: {repr(e)}")
            return


def download_all_photos(targets):
    """
    Downloads all photos given a list of targets which are tuples containing the url, directory, and filename.

    Parameters
    ----------
    targets : list of tuple of str
        Contains tuples that store the url, directory, filename, and resolution (will be None to get original photo resolution) of the desired photos to be downloaded in that order.
    """
    expectedNumParams = 4
    if pd.isna(targets):
        warnings.warn("Targets was none")
    else:
        for target in targets:
            if (type(target) is tuple) and len(target) == expectedNumParams:
                download_photo(*target)
            else:
                warnings.warn(f"Target incorrectly formatted: {target}")


def _format_param_name(name: str):
    if pd.isna(name):
        return None
    return (
        "".join(s.capitalize() + " " for s in name.split("_"))
        .replace("Photo", "")
        .strip()
    )


# Constructs Photo Name using given included fields and additional information
def _build_photo_name(
    protocol, photo_id, name_fields, include_in_name=[], additional_name_stem=""
):
    valid_protocols = ["lc_", "mhm_"]
    if not protocol or protocol not in valid_protocols:
        warnings.warn("Invalid protocol")
        return None
    name = protocol
    if additional_name_stem and additional_name_stem != "":
        name += f"{additional_name_stem}_"

    if include_in_name:
        for field in list(include_in_name):
            if field in set(name_fields):
                name += f"{name_fields[field]}_"

    name += f"{photo_id}.png"
    name = remove_bad_characters(name)
    return name


def _get_mosquito_classification(genus, species):
    classification = genus
    if pd.isna(classification):
        classification = "None"
    elif not pd.isna(species):
        classification = f"{classification} {species}"
    return classification


def _warn_num_invalid_photos(num_invalid_photos: dict):
    if sum(num_invalid_photos.values()) > 0:
        msg = f"Skipped {sum(num_invalid_photos.values())} invalid photos: "
        msg += str(num_invalid_photos)
        warnings.warn(msg)


def get_mhm_download_targets(
    mhm_df,
    directory,
    latitude_col="mhm_Latitude",
    longitude_col="mhm_Longitude",
    watersource_col="mhm_WaterSource",
    date_col="mhm_measuredDate",
    id_col="mhm_MosquitoHabitatMapperId",
    genus_col="mhm_Genus",
    species_col="mhm_Species",
    larvae_photo="mhm_LarvaFullBodyPhotoUrls",
    watersource_photo="mhm_WaterSourcePhotoUrls",
    abdomen_photo="mhm_AbdomenCloseupPhotoUrls",
    include_in_name=[],
    additional_name_stem="",
    resolution=None,
):
    """
    Generates mosquito habitat mapper targets to download

    Parameters
    ----------
    mhm_df : pd.DataFrame
        Mosquito Habitat Mapper Data
    directory : str
        The directory to save the photos
    latitude_col : str, default="mhm_Latitude"
        The column name of the column that contains the Latitude
    longitude_col : str, default="mhm_Longitude"
        The column name of the column that contains the Longitude
    watersource_col : str, default = "mhm_WaterSource"
        The column name of the column that contains the watersource
    date_col : str, default = "mhm_measuredDate"
        The column name of the column that contains the measured date
    id_col : str, default = "mhm_MosquitoHabitatMapperId"
        The column name of the column that contains the mosquito habitat mapper id
    genus_col : str, default = "mhm_Genus"
        The column name of the column that contains the genus
    species_col : str, default = "mhm_Species"
        The column name of the column that contains the species
    larvae_photo : str, default = "mhm_LarvaFullBodyPhotoUrls"
        The column name of the column that contains the larvae photo urls. If not specified, the larvae photos will not be included.
    watersource_photo : str, default = "mhm_WaterSourcePhotoUrls"
        The column name of the column that contains the watersource photo urls. If not specified, the larvae photos will not be included.
    abdomen_photo : str, default = "mhm_AbdomenCloseupPhotoUrls"
        The column name of the column that contains the abdomen photo urls. If not specified, the larvae photos will not be included.
    include_in_name : list of str, default=[]
        A list of column names to include into the downloaded photo names. The order of items in this list is maintained in the outputted name
        Accepted Included Names include:
            * `url_type` -- Type of photo (e.g. Watersource, Larvae, Abdomen)
            * `watersource` -- Watersource for the observed mosquito habitat
            * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
            * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
            * `date_str` -- Date Range expressed as a String
            * `mhm_id` -- Unique ID for a MHM observation
            * `classification` -- Mosquito classification (or `"None"` if no classification available)
    additional_name_stem : str, default=""
        Additional custom information the user can add to the name.
    resolution : tuple of int, default = None
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image. If the resolution is None, the original resolution of the photo is downloaded.

    Returns
    -------
    set of tuple of str
        Contains the (url, directory, and filename) for each desired mosquito habitat mapper photo
    """
    arguments = locals()
    targets = set()
    num_invalid_photos = {
        "invalid_URL": 0,
        "rejected": 0,
        "pending": 0,
        "bad_photo_id": 0,
    }

    def get_photo_args(
        url_entry,
        url_type,
        latitude,
        longitude,
        watersource,
        date,
        mhm_id,
        genus,
        species,
    ):
        if pd.isna(url_entry):
            return

        urls = url_entry.split(";")
        date_str = pd.to_datetime(str(date)).strftime("%Y-%m-%d")

        for url in urls:
            if not pd.isna(url) and "https" in url:
                photo_id = get_globe_photo_id(url)

                name_fields = {
                    "url_type": url_type,
                    "watersource": watersource,
                    "latitude": round(latitude, 5),
                    "longitude": round(longitude, 5),
                    "date_str": date_str,
                    "mhm_id": mhm_id,
                    "classification": _get_mosquito_classification(genus, species),
                }

                # Checks photo_id is valid
                if not pd.isna(photo_id) and int(photo_id) >= 0:
                    protocol = "mhm_"
                    name = _build_photo_name(
                        protocol,
                        photo_id,
                        name_fields,
                        include_in_name,
                        additional_name_stem,
                    )
                    targets.add((url, directory, name, resolution))
                else:
                    num_invalid_photos["bad_photo_id"] += 1
            elif not pd.isna(url) and "rejected" in url:
                num_invalid_photos["rejected"] += 1
            elif not pd.isna(url) and "pending" in url:
                num_invalid_photos["pending"] += 1
            else:
                num_invalid_photos["invalid_URL"] += 1

    photo_locations = {k: v for k, v in arguments.items() if "photo" in k}
    for param_name, column_name in photo_locations.items():
        if column_name:
            get_mosquito_args = np.vectorize(get_photo_args)
            get_mosquito_args(
                mhm_df[column_name].to_numpy(),
                _format_param_name(param_name),
                mhm_df[latitude_col].to_numpy(),
                mhm_df[longitude_col].to_numpy(),
                mhm_df[watersource_col].to_numpy(),
                mhm_df[date_col],
                mhm_df[id_col].to_numpy(),
                mhm_df[genus_col].to_numpy(),
                mhm_df[species_col].to_numpy() if species_col else "",
            )
    _warn_num_invalid_photos(num_invalid_photos)
    return targets


def download_mhm_photos(
    mhm_df,
    directory,
    latitude_col="mhm_Latitude",
    longitude_col="mhm_Longitude",
    watersource_col="mhm_WaterSource",
    date_col="mhm_measuredDate",
    id_col="mhm_MosquitoHabitatMapperId",
    genus_col="mhm_Genus",
    species_col="mhm_Species",
    larvae_photo="mhm_LarvaFullBodyPhotoUrls",
    watersource_photo="mhm_WaterSourcePhotoUrls",
    abdomen_photo="mhm_AbdomenCloseupPhotoUrls",
    include_in_name=[],
    additional_name_stem="",
    resolution=None,
):
    """
    Downloads mosquito habitat mapper photos

    Parameters
    ----------
    mhm_df : pd.DataFrame
        Mosquito Habitat Mapper Data
    directory : str
        The directory to save the photos
    latitude_col : str, default="mhm_Latitude"
        The column name of the column that contains the Latitude
    longitude_col : str, default="mhm_Longitude"
        The column name of the column that contains the Longitude
    watersource_col : str, default = "mhm_WaterSource"
        The column name of the column that contains the watersource
    date_col : str, default = "mhm_measuredDate"
        The column name of the column that contains the measured date
    id_col : str, default = "mhm_MosquitoHabitatMapperId"
        The column name of the column that contains the mosquito habitat mapper id
    genus_col : str, default = "mhm_Genus"
        The column name of the column that contains the genus
    species_col : str, default = "mhm_Species"
        The column name of the column that contains the species
    larvae_photo : str, default = "mhm_LarvaFullBodyPhotoUrls"
        The column name of the column that contains the larvae photo urls. If not specified, the larvae photos will not be included.
    watersource_photo : str, default = "mhm_WaterSourcePhotoUrls"
        The column name of the column that contains the watersource photo urls. If not specified, the larvae photos will not be included.
    abdomen_photo : str, default = "mhm_AbdomenCloseupPhotoUrls"
        The column name of the column that contains the abdomen photo urls. If not specified, the larvae photos will not be included.
    include_in_name : list of str, default=[]
        A list of column names to include into the downloaded photo names. The order of items in this list is maintained in the outputted name list of column names to include into the downloaded photo names
        Accepted Included Names include:
            * `url_type` -- Type of photo (e.g. Watersource, Larvae, Abdomen)
            * `watersource` -- Watersource for the observed mosquito habitat
            * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
            * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
            * `date_str` -- Date Range expressed as a String
            * `mhm_id` -- Unique ID for a MHM observation
            * `classification` -- Mosquito classification (or `"None"` if no classification available)
    additional_name_stem : str, default=""
        Additional custom information the user can add to the name.
    resolution : tuple of int, default = None
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image. If the resolution is None, the original resolution of the photo is downloaded.

    Returns
    -------
    set of tuple of str
        Contains the (url, directory, and filename) for each desired mosquito habitat mapper photo
    """
    targets = get_mhm_download_targets(**locals())
    download_all_photos(targets)
    return targets


def get_lc_download_targets(
    lc_df,
    directory,
    latitude_col="lc_Latitude",
    longitude_col="lc_Longitude",
    date_col="lc_measuredDate",
    id_col="lc_LandCoverId",
    up_photo="lc_UpwardPhotoUrl",
    down_photo="lc_DownwardPhotoUrl",
    north_photo="lc_NorthPhotoUrl",
    south_photo="lc_SouthPhotoUrl",
    east_photo="lc_EastPhotoUrl",
    west_photo="lc_WestPhotoUrl",
    include_in_name=[],
    additional_name_stem="",
    resolution=None,
):
    """
    Generates landcover targets to download

    Parameters
    ----------
    lc_df : pd.DataFrame
        Cleaned and Flagged Landcover Data
    directory : str
        The directory to save the photos
    latitude_col : str, default="lc_Latitude"
        The column of the column that contains the Latitude
    longitude_col : str, default="lc_Longitude"
        The column of the column that contains the Longitude
    date_col : str, default="lc_measuredDate"
        The column name of the column that contains the measured date
    id_col : str, default="lc_LandCoverId"
        The column name of the column that contains the landcover id
    up_photo : str, default = "lc_UpwardPhotoUrl"
        The column name of the column that contains the upward photo urls. If not specified, these photos will not be included.
    down_photo : str, default = "lc_DownwardPhotoUrl"
        The column name of the column that contains the downward photo urls. If not specified, these photos will not be included.
    north_photo : str, default = "lc_NorthPhotoUrl"
        The column name of the column that contains the north photo urls. If not specified, these photos will not be included.
    south_photo : str, default = "lc_SouthPhotoUrl"
        The column name of the column that contains the south photo urls. If not specified, these photos will not be included.
    east_photo : str, default = "lc_EastPhotoUrl"
        The column name of the column that contains the east photo urls. If not specified, these photos will not be included.
    west_photo : str, default = "lc_WestPhotoUrl"
        The column name of the column that contains the west photo urls. If not specified, these photos will not be included.
    include_in_name : list of str, default=[]
        A list of column names to include into the downloaded photo names. The order of items in this list is maintained in the outputted name
        Accepted Included Names include:
            * `direction` -- Direction where the photo was taken (e.g. North, South, East, West, Up, Down)
            * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
            * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
            * `date_str` -- Date Range expressed as a String
            * `lc_id` -- Unique ID for a LC observation
    additional_name_stem : str, default=""
        Additional custom information the user can add to the name.
    resolution : tuple of int, default = None
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image. If the resolution is None, the original resolution of the photo is downloaded.

    Returns
    -------
    set of tuple of str
        Contains the (url, directory, and filename) for each desired land cover photo
    """
    arguments = locals()
    targets = set()
    num_invalid_photos = {
        "invalid_URL": 0,
        "rejected": 0,
        "pending": 0,
        "bad_photo_id": 0,
    }

    def get_photo_args(url, latitude, longitude, direction, date, lc_id):
        if not pd.isna(url) and "https" in url:
            date_str = pd.to_datetime(str(date)).strftime("%Y-%m-%d")
            photo_id = get_globe_photo_id(url)

            name_fields = {
                "direction": direction,
                "latitude": round(latitude, 5),
                "longitude": round(longitude, 5),
                "date_str": date_str,
                "lc_id": lc_id,
            }

            if not pd.isna(photo_id) and int(photo_id) >= 0:
                protocol = "lc_"
                name = _build_photo_name(
                    protocol,
                    photo_id,
                    name_fields,
                    include_in_name,
                    additional_name_stem,
                )
                targets.add((url, directory, name, resolution))
            else:
                num_invalid_photos["bad_photo_id"] += 1
        elif not pd.isna(url) and "rejected" in url:
            num_invalid_photos["rejected"] += 1
        elif not pd.isna(url) and "pending" in url:
            num_invalid_photos["pending"] += 1
        else:
            num_invalid_photos["invalid_URL"] += 1

    photo_locations = {k: v for k, v in arguments.items() if "photo" in k}
    for param_name, column_name in photo_locations.items():
        if column_name:
            get_lc_photo_args = np.vectorize(get_photo_args)
            get_lc_photo_args(
                lc_df[column_name].to_numpy(),
                lc_df[latitude_col].to_numpy(),
                lc_df[longitude_col].to_numpy(),
                _format_param_name(param_name),
                lc_df[date_col],
                lc_df[id_col].to_numpy(),
            )
    _warn_num_invalid_photos(num_invalid_photos)
    return targets


def download_lc_photos(
    lc_df,
    directory,
    latitude_col="lc_Latitude",
    longitude_col="lc_Longitude",
    date_col="lc_measuredDate",
    id_col="lc_LandCoverId",
    up_photo="lc_UpwardPhotoUrl",
    down_photo="lc_DownwardPhotoUrl",
    north_photo="lc_NorthPhotoUrl",
    south_photo="lc_SouthPhotoUrl",
    east_photo="lc_EastPhotoUrl",
    west_photo="lc_WestPhotoUrl",
    include_in_name=[],
    additional_name_stem="",
    resolution=None,
):
    """
    Downloads Landcover photos for landcover data.

    Parameters
    ----------
    lc_df : pd.DataFrame
        Cleaned and Flagged Landcover Data
    directory : str
        The directory to save the photos
    latitude_col : str, default="lc_Latitude"
        The column of the column that contains the Latitude
    longitude_col : str, default="lc_Longitude"
        The column of the column that contains the Longitude
    date_col : str, default="lc_measuredDate"
        The column name of the column that contains the measured date
    id_col : str, default="lc_LandCoverId"
        The column name of the column that contains the landcover id
    up_photo : str, default = "lc_UpwardPhotoUrl"
        The column name of the column that contains the upward photo urls. If not specified, these photos will not be included.
    down_photo : str, default = "lc_DownwardPhotoUrl"
        The column name of the column that contains the downward photo urls. If not specified, these photos will not be included.
    north_photo : str, default = "lc_NorthPhotoUrl"
        The column name of the column that contains the north photo urls. If not specified, these photos will not be included.
    south_photo : str, default = "lc_SouthPhotoUrl"
        The column name of the column that contains the south photo urls. If not specified, these photos will not be included.
    east_photo : str, default = "lc_EastPhotoUrl"
        The column name of the column that contains the east photo urls. If not specified, these photos will not be included.
    west_photo : str, default = "lc_WestPhotoUrl"
        The column name of the column that contains the west photo urls. If not specified, these photos will not be included.
    include_in_name : list of str, default=[]
        A list of column names to include into the downloaded photo names. The order of items in this list is maintained in the outputted name
        Accepted Included Names include:
            * `direction` -- Direction where the photo was taken (e.g. North, South, East, West, Up, Down)
            * `latitude` -- GPS Latitude Coordinate (rounded to 5 decimal places)
            * `longitude` -- GPS Longitude Coordinate (rounded to 5 decimal places)
            * `date_str` -- Date Range expressed as a String
            * `lc_id` -- Unique ID for a LC observation
    additional_name_stem : str, default=""
        Additional custom information the user can add to the name.
    resolution : tuple of int, default = None
        The image resolution in width x height. e.g. (1920, 1080) for a 1080p image. If the resolution is None, the original resolution of the photo is downloaded.

    Returns
    -------
    set of tuple of str
        Contains the (url, directory, and filename) for each desired land cover photo
    """
    targets = get_lc_download_targets(**locals())
    download_all_photos(targets)
    return targets



import logging
import math

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def plot_freq_bar(df, protocol, column, title, plot_type="bar", log_scale=False):
    """
    Plots the frequency of different values of a column across the dataset

    Parameters
    ----------
    df : pd.DataFrame
      The DataFrame containing the desired data.
    protocol : str
      The Protocol of the DataFrame (e.g. Mosquito Habitat Mapper or Land Cover)
    column : str
      The name of the column that the graph will plot.
    title : str
      The title of the graph
    plot_type : str, default="bar"
      The type of graph that will be used to visually compare the data.
    log_scale : bool, default=False
      If a log scale will be used for the data. If true, all the values will be in Log 10.
    """
    col_vals = pd.DataFrame()
    col_vals["Count"] = df.groupby(column).size()
    col_vals.reset_index(inplace=True)
    col_vals.columns = ["index", "Count"]
    if len(col_vals) <= 1:  # pragma: no cover
        logging.warning(
            f"There is only one value for this column: {col_vals['index'][0]}: {col_vals['Count'][0]}"
        )
    plt.figure(figsize=(10, 6))
    ylabel = "Frequency"
    title = f"{protocol} - {title}"
    if log_scale:
        col_vals["Count"] = pd.Series([math.log10(val) for val in col_vals["Count"]])
        ylabel += " (Log Scale)"
        title += " (Log Scale)"
    if plot_type == "line":  # pragma: no cover
        plt.plot(col_vals["index"], col_vals["Count"], color="lightblue")
    else:
        plt.bar(col_vals["index"], col_vals["Count"], color="lightblue")

    plt.title(title)
    plt.xlim(left=-0.5)
    plt.xlabel(f"{column} Values")
    plt.ylabel(ylabel)


def multiple_bar_graph(df, protocol, cols, title, log_scale=False):
    """
    Plots the frequency of different values of a column across the dataset alongside eachother.

    Parameters
    ----------
    df : pd.DataFrame
      The DataFrame containing the desired data.
    protocol : str
      The Protocol of the DataFrame (e.g. Mosquito Habitat Mapper or Land Cover)
    cols : list str
      The names of the columns that the graph will plot.
    title : str
      The title of the graph
    log_scale : bool, default=False
      If a log scale will be used for the data. If true, all the values will be in Log 10.
    """

    def create_summary_df(cols):
        """
        data = df
        x = [0,6]
        hue = photoType: photocount, rejectedcount, etc
        """
        photo_summary = pd.DataFrame()
        for name in cols:
            photo_summary[name] = df[name].value_counts()
            # print(type(photo_summary[name]))
        photo_summary.sort_index(inplace=True)
        photo_summary = photo_summary.reset_index()
        # print(photo_summary)

        category_counts = []

        for name in cols:
            for j in range(len(photo_summary)):
                count = photo_summary[name][j]
                if log_scale:
                    count = math.log10(count)
                new_row = {
                    "index": photo_summary["index"][j],
                    "category": name,
                    "count": count,
                }
                category_counts.append(new_row)
        category_counts = pd.DataFrame(category_counts)
        return category_counts

    category_counts = create_summary_df(cols)
    plt.figure(figsize=(10, 6))

    title = f"{protocol} -- {title}"
    ylabel = "Frequency"
    if log_scale:
        ylabel += " (Log Scale)"
        title += " (Log Scale)"

    ax = sns.barplot(
        data=category_counts,
        x="index",
        y="count",
        hue="category",
        palette=[
            "#377eb8",
            "#ff7f00",
            "#4daf4a",
            "#f781bf",
            "#a65628",
            "#984ea3",
            "#999999",
            "#e41a1c",
            "#dede00",
        ],
    )
    ax.set_xlabel("Photo Count")
    ax.set_ylabel(ylabel)

    ax.set_title(title)
    plt.legend(loc="upper right")


def plot_int_distribution(df, col_name, title_name):
    """
    Plots the frequency of different integer values of a column across a cleaned dataset

    Parameters
    ----------
    df : pd.DataFrame
      The DataFrame containing the desired data.
    col_name : str
      The name of the column that the graph will plot.
    title_name : str
      The name of the column as you would like to have as the title (e.g. mhm_Genus could be just Genus in the title)
    """
    df = df.copy()
    df[col_name] = df[col_name].replace(-9999, -5)

    counts = df.groupby(col_name).size()
    title = f"{title_name} Distribution (with Null)"
    plt.figure(figsize=(10, 5))
    plt.title(title)
    plt.ylabel(f"{title_name} Entries  (Log Scale)")
    plt.yscale("log")
    plt.bar(counts.index, counts, color="#b30000")


def completeness_histogram(df, protocol, completeness_col, completeness_type):
    """
    Plots a histogram of the completeness score distribution.

    Parameters
    ----------
    df : pd.DataFrame
      The DataFrame containing the desired data.
    protocol : str
      The name of the protocol that the graph will plot.
    completeness_col : str
      The column containing the desired completeness metric
    completness_type : str
      The type of completeness score (Sub or Cumulative)
    """
    plt.figure(figsize=(10, 4))
    title = f"{protocol} -- {completeness_type} Completeness Scores Frequency Histogram"
    plt.title(title)
    plt.hist(df[completeness_col], color="pink", label=completeness_type)
    plt.xlabel("Scores")
    plt.ylabel("Count")


def save_stored_plots():
    """
    Saves any generated graphs currently stored by the pyplot object.
    """
    for num in plt.get_fignums():
        plt.figure(num)
        title = plt.gca().get_title()
        plt.savefig(f"{title}.png")
